from ctypes import *
from multiprocessing import Array, Pipe, Queue, shared_memory
import os
import queue
import time

import numpy as np
from expam.database import TIMEOUT, DataTypeConfig, FileLocationConfig, expam_dtypes
from expam.database.config import JSONConfig, load_database_config
from expam.database.db import TablesDb
from expam.database.shm import ResizableSharedMemory
from expam.logger import new_logger
from expam.process import COMMAND_CODES
from expam.process.jobworker import JobWorker


class ControlCenter:
    def __init__(
        self, logging_dir, group_name=None, workers=None, child_statuses=None, phases=None,
        phase_queues=None, child_queues=None, children=None, processors=None, transitions=None,
        timeout=TIMEOUT
):
        self.group_name = group_name
        self.workers = [] if workers is None else list(workers)

        self.error = False  # Error state.
        self._child_statuses = child_statuses  # Shared array.

        self.phases = [] if phases is None else list(phases)
        self.phase_queues = {} if phase_queues is None else dict(phase_queues)
        self.child_queues = child_queues

        # Connection objects.
        self._children = [] if children is None else list(children)

        # User defined methods.
        self._processors = {} if processors is None else dict(processors)
        self._transitions = {} if transitions is None else dict(transitions)
        self._shutdown = self.default_shutdown

        self._timeout = timeout

        # Logging to disk.
        self.logging_dir = logging_dir
        self.logger = new_logger(self.logging_dir, str(os.getpid()) + "_main")

    @property
    def phase(self):
        if not self.phases:
            return None

        else:
            return self.phases[0]

    def children_have_obeyed(self, *command_codes):
        if self._child_statuses is None:
            return True

        return all([
            child_status in command_codes
            for child_status in self._child_statuses
        ])

    @property
    def any_children_active(self):
        if self._child_statuses is None:
            return False

        return any([
            child_status == COMMAND_CODES["ACTIVE"]
            for child_status in self._child_statuses
        ])

    @property
    def children_are_dead(self):
        return self.children_have_obeyed(COMMAND_CODES["DEAD"])

    @classmethod
    def from_dict(cls, logging_dir, config):
        def check_int(value):
            if not isinstance(value, int):
                return False

            return True

        if not isinstance(config, dict):
            raise TypeError("Can only be instantiated using dictionary not %s!" % type(config))

        control_center = ControlCenter(logging_dir)

        required_args = [
            "phases",
            "layers",
            "name",
        ]
        optional_args = [
            "timeout",
        ]

        # Set optional arguments first.
        for arg in optional_args:
            if arg in config:

                if arg == "timeout":
                    if type(config[arg]) not in {float, int}:
                        raise ValueError("Parameter must be float or int not %s!" % type(config[arg]))

                    else:
                        control_center._timeout = config["timeout"]

        # Set required arguments.
        for arg in required_args:
            if arg not in config:
                raise ValueError("Required arg %s not in configuration!" % arg)

            else:
                if arg == "phases":
                    control_center.phases = list(config["phases"])

                elif arg == "layers":
                    layers_required_args = {
                        "class",
                        "parent",
                        "child",
                        "n"
                    }
                    current_layer = config["layers"]

                    if not isinstance(current_layer, dict) and "parent" in current_layer:
                        raise ValueError("Incorrect layer configuration!")

                    # Ensure that top layer is directly connected to the control center.
                    if current_layer["parent"] is not ControlCenter:
                        raise TypeError("Top level worker must be connected to ControlCenter not %s!"
                                        % current_layer["parent"])

                    parents = [control_center]

                    at_bottom = False
                    while not at_bottom:

                        if not isinstance(current_layer, dict):
                            raise ValueError("Layer configuration must be dictionary not %s!" % type(current_layer))

                        # Validate arguments.
                        for layer_arg in layers_required_args:
                            if layer_arg not in current_layer:
                                raise ValueError("Missing %s argument required for layer creation!" % layer_arg)

                            elif layer_arg == "n":
                                if not check_int(current_layer["n"]):
                                    raise TypeError("Layer parameter n must be int not %s!" % type(current_layer["n"]))

                            elif layer_arg == "class":
                                if not issubclass(current_layer["class"], JobWorker):
                                    raise TypeError("Process %s must be subclass of JobWorker!"
                                                    % str(current_layer["class"].__name__))

                            elif layer_arg == "child":
                                if current_layer["child"] is None:
                                    at_bottom = True

                        parents = control_center.add_layer(current_layer, parents)
                        current_layer = current_layer["child"]

                elif arg == "name":
                    control_center.group_name = str(config["name"])

        return control_center

    def add_layer(self, current_layer, parents):

        n_children = current_layer["n"]
        n_parents = len(parents)

        def make_shared_array():
            nonlocal n_children

            arr = Array(c_int8, n_children, lock=False)  # Only edited by children.
            for i in range(n_children):
                arr[i] = COMMAND_CODES["ACTIVE"]

            return arr

        def initialise_workers(statuses, timeout):
            nonlocal current_layer, n_children

            # Class of worker.
            worker = current_layer["class"]

            # Custom to initialise the worker.
            if "class_args" in current_layer:

                init_args = [
                    {
                        arg: value
                        if not (isinstance(value, tuple) and len(value) == n_children) else value[cid]
                        for arg, value in current_layer["class_args"].items()
                    }
                    for cid in range(n_children)
                ]

            else:
                init_args = [{} for _ in range(n_children)]

            # Insert default arguments.
            for cid in range(n_children):
                default_args = {
                    "status_index": cid,
                    "statuses": statuses,
                    "child_statuses": None,
                    "timeout": timeout,
                    "phases": self.phases,
                }

                init_args[cid] = {
                    **init_args[cid],
                    **default_args
                }

            return [
                worker(**init_args[cid])
                for cid in range(n_children)
            ]

        def make_pipes():
            nonlocal n_parents, n_children

            parent_cons = [[] for _ in range(n_parents)]
            child_cons = [[] for _ in range(n_children)]

            for i in range(n_parents):
                for j in range(n_children):
                    parent_con, child_con = Pipe()

                    parent_cons[i].append(parent_con)
                    child_cons[j].append(child_con)

            return parent_cons, child_cons

        def make_queues():
            nonlocal n_children

            return [
                Queue()
                for _ in range(n_children)
            ]

        current_statuses = make_shared_array()
        parent_cons, child_cons = make_pipes()
        work_queues = make_queues()

        for i, parent in enumerate(parents):
            parent._child_statuses = current_statuses
            parent._children = parent_cons[i]
            parent.child_queues = work_queues

        workers = initialise_workers(
            current_statuses,
            self._timeout
        )

        for i, worker in enumerate(workers):
            # Set communication routes.
            worker._fellow_statuses = current_statuses
            worker._parents = child_cons[i]
            worker.work_queue = work_queues[i]

        self.workers.append(workers)

        return workers

    def set_methods(self, methods_dict):

        if not isinstance(methods_dict, dict):
            raise TypeError("Config must be dictionary not %s!" % str(type(methods_dict)))

        for key, value in methods_dict.items():
            if value is None:
                continue

            if key == "shutdown":

                if not callable(value):
                    raise TypeError("Shutdown object must be callable!")

                self._shutdown = value

            elif key in self.phases:

                if not isinstance(value, dict):
                    raise TypeError("Method declaration must be dictionary not %s!"
                                    % str(type(value)))

                for method_name, method in value.items():
                    self.set_method(
                        method_name=method_name,
                        phase=key,
                        method=method
                    )

            else:
                raise ValueError("Unknown key in class config %s!" % key)

    def set_method(self, method_name, phase, method):
        methods = {
            "processor": self._processors,
            "transition": self._transitions,
        }

        if method_name not in methods:
            raise ValueError("Invalid method name %s!" % method_name)

        if method is not None:
            methods[method_name][phase] = method

    def default_shutdown(self):
        self.log("Control center shutting down.")

    def run(self):
        self.start()

        try:

            # Phase loop.
            while self.phases:
                phase = self.phases[0]

                try:
                    self.log("Main filling queue at phase %s..." % self.phase)

                    # Processing loop.
                    while True:
                        # Check children for errors and any custom processing protocols.
                        self.process()

                        for child_queue in self.child_queues:
                            next_job = self.phase_queues[phase].get(block=True, timeout=self._timeout)

                            child_queue.put(next_job)

                except queue.Empty:
                    self.log("Exiting loop from MAIN...")

                except KeyError:

                    # Check there is work for this phase.
                    if phase not in self.phase_queues:
                        pass

                    else:
                        raise

                # Processing finished.
                self.log("Telling children to transition...")
                # Finalise data and transition to next phase.
                self.tell_children(
                    (
                        COMMAND_CODES["SLEEP"],
                        None
                    )
                )

                # Wait for response.
                while not self.children_have_obeyed(COMMAND_CODES["SLEEP"], COMMAND_CODES["DEAD"]):
                    self.process()

                # Transition now that children are transitioning.
                self.transition()  # Progresses current phase.

            # Shutdown phase.
            # Wait for processes to finish.

            while not self.children_have_obeyed(COMMAND_CODES["DEAD"]):
                self.check_children()

                self.timeout()

        except Exception as e:

            # Kill children before raising error.
            # Propagate error downward.
            self.finalise_error(
                -1,
                COMMAND_CODES["ERROR"],
                e
            )

            raise e

        finally:

            self.shutdown()

    def start(self):
        for layer in self.workers:
            for worker in layer:
                worker.start()

    def shutdown(self):
        # Default shutdown procedure.
        self._shutdown()

        # Gather processes.
        self.join()

    def process(self):
        self.check_children()

        if self.phase in self._processors:
            self._processors[self.phase]()

    def transition(self):
        self.log("Main transitioning...")

        if self.phase in self._transitions:
            self._transitions[self.phase]()

        self.phases.pop(0)
        self.log("Main now at phase %s." % ("end" if not self.phases else self.phase))

        # Wake up children and begin next phase.
        self.wake_children()
        self.log("Children are now awake, beginning next phase...")

    def wake_children(self):
        for cid in range(len(self._child_statuses)):
            self._child_statuses[cid] = COMMAND_CODES["ACTIVE"]

    def join(self):
        # Check that my queues are empty.
        all_queues = [*self.child_queues, *[phase_queue for phase_queue in self.phase_queues.values()]]

        for child_queue in all_queues:
            while True:
                try:
                    _ = child_queue.get(block=True, timeout=self._timeout)

                except queue.Empty:
                    break

        # Clean up workers.
        for layer in self.workers:
            for worker in layer:
                worker.join()

    def kill(self):
        for cid in range(len(self._children)):
            self.tell_child(
                cid=cid,
                message=(
                    COMMAND_CODES["DEAD"],
                    None
                )
            )

    def tell_children(self, message):
        for cid in range(len(self._children)):
            self.tell_child(cid, message)

    def tell_child(self, cid, message):
        self.child_queues[cid].put(message)

    def check_children(self):
        for cid in range(len(self._children)):
            self.check_child(cid)

    def check_child(self, cid):
        if self._children[cid].poll(self._timeout):
            err_code, err = self._children[cid].recv()
            self._children[cid].send(True)  # Read receipt.

            self.log("Error received at main!")

            # Propagate error downward.
            self.finalise_error(cid, err_code, err)

            # Processes should have died if an error has been raised.
            self.join()

            print("\n" + "-" * 30 + "\n  Exception raised to MAIN:\n" + "-" * 30)
            raise err

    def finalise_error(self, eid, code, err):
        self.log("MAIN propagating error...")
        self.error = True

        for cid in range(len(self._children)):
            if cid != eid:
                self._children[cid].send((code, err))

        # Wait for children to die.
        while not self.children_are_dead:
            self.timeout()

    def load_job(self, phase, job):
        if phase not in self.phases:
            raise ValueError("Undefined job phase %s!" % phase)

        # Initialise new phase.
        if phase not in self.phase_queues:
            self.phase_queues[phase] = Queue()

        self.phase_queues[phase].put(
            (
                COMMAND_CODES["ACTIVE"],
                job
            )
        )

    def timeout(self):
        time.sleep(self._timeout)

    def log(self, msg):
        self.logger.info(msg)


class ExpamProcesses(ControlCenter):
    def __init__(self, logging_dir, cons, out_dir, dtype_config: DataTypeConfig = expam_dtypes, *args, **kwargs):
        super(ExpamProcesses, self).__init__(logging_dir, *args, **kwargs)

        # Define non-hierarchical connection to children.
        self.cons = cons
        self.out_dir = out_dir

        self.dtypes: DataTypeConfig = dtype_config
        self.database_config: FileLocationConfig = load_database_config(out_dir)

        # Current UnionWorker array sizes.
        self.arr_sizes = {}

    @classmethod
    def from_method_dict(cls, connections, out_dir, logging_dir, config):

        # Configure base class from configuration dictionary.
        base_center = super(ExpamProcesses, cls).from_dict(logging_dir, config)

        base_arguments = {
            "group_name": "group_name",
            "workers": "workers",
            "child_statuses": "_child_statuses",
            "phases": "phases",
            "phase_queues": "phase_queues",
            "child_queues": "child_queues",
            "children": "_children",
            "processors": "_processors",
            "transitions": "_transitions",
            "timeout": "_timeout",
        }
        base_attributes = {
            attr: getattr(base_center, attr_reference)
            for attr, attr_reference in base_arguments.items()
        }

        # Create child class from this instance.
        control_center = ExpamProcesses(
            logging_dir=logging_dir,
            cons=connections,
            out_dir=out_dir,
            **base_attributes
        )

        # Initialise special methods.
        control_center.set_methods(
            {  # Custom methods for non-hierarchical communication between layers.
                "import": {
                    "processor": control_center.union_checker,
                    "transition": control_center.union_checker,
                },
                "map": {
                    "processor": None,
                    "transition": None,
                },
                "shutdown": control_center.union_saver,
            }
        )

        return control_center

    def union_checker(self):
        for i, con in enumerate(self.cons):
            if con.poll(self._timeout):

                cmd, data = con.recv()

                if cmd == "shm_extend":
                    shm_name, size, create = data

                    if shm_name is None:
                        # Create an allocation.
                        shm_all = shared_memory.SharedMemory(size=size, create=create)

                        shm_name = shm_all.name
                        shm_all.close()

                    else:
                        # Resize a current allocation.
                        shm_all = ResizableSharedMemory(shm_name)

                        shm_all.resize(size)
                        shm_all.close()

                    con.send(shm_name)

                elif cmd == "keys_size":
                    keys_shape = data

                    # Update known size.
                    self.arr_sizes[i] = keys_shape

                    # Set map size.
                    map_shape = keys_shape[:1]

                    # Create a store for its values.
                    shm_values_size = np.product(map_shape) * self.dtypes.map_dtype_size
                    shm_all = shared_memory.SharedMemory(size=shm_values_size, create=True)

                    shm_name = shm_all.name
                    shm_all.close()

                    con.send(shm_name)

                else:
                    raise ValueError("Unrecognised command %s!" % cmd)

    def union_saver(self):
        if not self.error:

            # Calculate combined size.
            final_shape = tuple(
                fn([shape[i] for shape in self.arr_sizes.values()])
                for i, fn in enumerate((sum, max))
            )
            n_cols = final_shape[1]

            # Initialise database on disk.
            db_url = self.database_config.database_file
            disk_array = TablesDb(
                db_url=db_url,
                shape=final_shape,
                keys_type=self.dtypes.keys_dtype,
                values_type=self.dtypes.map_dtype,
                create=True
            )
            disk_array.close()

            def add_tuples(*tuples):
                if tuples:
                    return tuple(
                        sum([tup[i] for tup in tuples])
                        for i in range(len(tuples[0]))
                    )

            # Tell each process to write to disk.
            current_shape = (0,) * n_cols

            for con in self.cons:

                while not con.poll(self._timeout):
                    # Don't cause hang due to un-propagated errors!
                    self.check_children()

                cmd, shape = con.recv()

                if cmd != "save":
                    raise ValueError("Unrecognised command %s!" % cmd)

                # Let process know where it comes in the file.
                con.send((current_shape[0], db_url))

                if shape is not None:
                    current_shape = add_tuples(current_shape, shape)

            # Save database parameters to config file.
            self.save_db_params(
                db_params={
                    "keys_shape": current_shape,
                    "keys_data_type": self.dtypes.keys_dtype_str,
                    "values_shape": current_shape[:1],
                    "values_data_type": self.dtypes.map_dtype_str,
                }
            )

    def save_db_params(self, db_params):
        conf = JSONConfig(url=self.database_config.conf)
        conf.set(**db_params)
        conf.save()




import math
import multiprocessing

import numpy as np
from expam.database import TIMEOUT, DataTypeConfig, expam_dtypes
from expam.database.db import TablesDb
from expam.ext.map import binarysearch_put
from expam.ext.sets import disjoint as make_disjoint, filler
from expam.database.shm import ResizableSharedMemory
from expam.logger import log_timer
from expam.process.jobworker import JobWorker


class UnionWorker(JobWorker):
    def __init__(
        self, main_con, lca_matrix, pile_size, logging_dir,
        lock_value, status_index, statuses, child_statuses,
        timeout=TIMEOUT, phases=None, dtype_config: DataTypeConfig = expam_dtypes
    ):
        super(UnionWorker, self).__init__(
            logging_dir, status_index, statuses, child_statuses,
            timeout=timeout, phases=phases
        )

        # Connection to main process.
        self.job_name = "union"
        self.con = main_con

        # For mapping LCAs in mapping stage.
        self.lca_matrix = lca_matrix

        # Piling parameters.
        self.in_allocations = {}
        self.in_arrays = {}
        self.pile = []

        if pile_size == "inf":
            self.pile_threshold = math.inf

        else:
            pile_size = self.is_int(pile_size)

            if pile_size is not False:
                self.pile_threshold = pile_size

            else:
                self.pile_threshold = 255  # Default.

        # Shared memory allocations.
        self.dtypes: DataTypeConfig = dtype_config
        self.shm_allocations = {
            "keys": {
                "array": None,
                "allocation": None,
                "shm_name": None,
                "size": None,
                "data_type": dtype_config.keys_dtype,
                "data_type_size": dtype_config.keys_dtype_size,
                "active": False,
                "write_offset": None,
            },
            "values": {
                "array": None,
                "allocation": None,
                "shm_name": None,
                "size": None,
                "data_type": dtype_config.map_dtype,
                "data_type_size": dtype_config.map_dtype_size,
                "active": False,
                "write_offset": None,
            }
        }

        self.lock_value = lock_value

        # Configure working states.
        self.init_methods_from_dict(
            {
                "import": {
                    "getter": None,
                    "processor": self.import_kmers,
                    "transition": self.finish_import,
                },
                "map": {
                    "getter": None,
                    "processor": self.map,
                    "transition": None,
                },
                "shutdown": self.save_shutdown
            }
        )

    @property
    def shm_keys(self):
        return self.shm_allocations["keys"]["array"]

    @shm_keys.setter
    def shm_keys(self, value):
        self.shm_allocations["keys"]["array"] = value

    @property
    def shm_values(self):
        return self.shm_allocations["values"]["array"]

    @shm_values.setter
    def shm_values(self, value):
        self.shm_allocations["values"]["array"] = value

    @property
    def shm_keys_allocation(self):
        return self.shm_allocations["keys"]["allocation"]

    @shm_keys_allocation.setter
    def shm_keys_allocation(self, value):
        self.shm_allocations["keys"]["allocation"] = value

    @property
    def shm_values_allocation(self):
        return self.shm_allocations["values"]["allocation"]

    @shm_values_allocation.setter
    def shm_values_allocation(self, value):
        self.shm_allocations["values"]["allocation"] = value

    @property
    def shm_keys_name(self):
        return self.shm_allocations["keys"]["shm_name"]

    @shm_keys_name.setter
    def shm_keys_name(self, value):
        self.shm_allocations["keys"]["shm_name"] = value

    @property
    def shm_values_name(self):
        return self.shm_allocations["values"]["shm_name"]

    @shm_values_name.setter
    def shm_values_name(self, value):
        self.shm_allocations["values"]["shm_name"] = value

    @property
    def shm_keys_size(self):
        return self.shm_allocations["keys"]["size"]

    @shm_keys_size.setter
    def shm_keys_size(self, value):
        self.shm_allocations["keys"]["size"] = value

    @property
    def shm_values_size(self):
        return self.shm_allocations["values"]["size"]

    @shm_values_size.setter
    def shm_values_size(self, value):
        self.shm_allocations["values"]["size"] = value

    @property
    def keys_data_type(self):
        return self.shm_allocations["keys"]["data_type"]

    @property
    def keys_data_type_size(self):
        return self.shm_allocations["keys"]["data_type_size"]

    @property
    def values_data_type(self):
        return self.shm_allocations["values"]["data_type"]

    @property
    def values_data_type_size(self):
        return self.shm_allocations["values"]["data_type_size"]

    @property
    def pile_length(self):
        return len(self.pile)

    @property
    def size(self):
        return sum([
            self.product(*kmers.shape, self.keys_data_type_size)
            for kmers in self.yield_kmers()
        ])

    @property
    def shape(self):
        return tuple(
            fn([kmers.shape[i] for kmers in self.yield_kmers()])
            for i, fn in enumerate((sum, max))
        )

    def yield_kmers(self):
        if self.shm_keys is not None:
            yield self.shm_keys

        for arr in self.pile:
            yield arr

    @staticmethod
    def is_int(string):
        try:
            value = float(string)
            int_value = int(value)

            if int_value != value:
                return False

            # Positive definite.
            if int_value < 1:
                return False

            return int_value

        except TypeError:
            return False

    @staticmethod
    def product(*args):

        total = 1
        for arg in args:
            total *= arg

        return total

    def import_kmers(self, data):
        window, shm_data = data
        kmers = self.view_kmers(window, shm_data)

        NO_KMERS_INFO = "No new kmers!"
        self.logger.info("Importing %d kmers at %d."
                         % (kmers.shape[0], self.pid))

        if kmers.shape[0] == 0:

            self.logger.info(NO_KMERS_INFO)

        else:

            # Make disjoint.
            kmers = self.disjoint(kmers)

            # If these kmers have no information to offer, continue.
            if kmers.shape[0] == 0:
                self.logger.info(NO_KMERS_INFO)

            else:
                self.pile.append(kmers)

                # Check if pile needs to be collapsed.
                if self.pile_length >= self.pile_threshold:
                    self.collapse()

                else:
                    self.logger.info("Process %d waiting this time around (%d/%s)."
                                     % (self.pid, self.pile_length, str(self.pile_threshold)))

    def view_kmers(self, window, data):
        start, end = window
        name, shape = data

        # Establish shared_memory connection.
        if name not in self.in_arrays:
            self.in_allocations[name] = multiprocessing.shared_memory.SharedMemory(name)
            self.in_arrays[name] = np.ndarray(
                shape=shape,
                dtype=self.dtypes.keys_dtype,
                buffer=self.in_allocations[name].buf
            )

        return (self.in_arrays[name])[start:end]

    @log_timer
    def disjoint(self, new_kmers):
        # Null case.
        n = new_kmers.shape[0]
        # Establish list of kmers against which to cull.
        comparisons = [self.shm_keys, *self.pile] if self.shm_keys is not None else self.pile

        # Remove any kmers that appear in the pile.
        n = make_disjoint([new_kmers, *comparisons])
        # Copy values into a new array.
        distinct_kmers = np.copy(new_kmers[:n])

        # Demonstrate completion.
        new_kmers[:] = 0

        return distinct_kmers

    def collapse(self):
        # Calculate new array parameters.
        size = self.size
        shape = self.shape

        # Remember old shape for filling.
        old_shape = 0 if self.shm_keys is None else self.shm_keys.shape[0]
        self.logger.info("Process %d sending extension request to shape (%d, %d)."
                         % (self.pid, shape[0], shape[1]))

        # Request extension to memory.
        create = self.shm_keys is None
        self.request_extension(size, create=create)

        # Remap memory extension.
        if self.shm_keys_allocation is None:
            self.shm_keys_allocation = ResizableSharedMemory(self.shm_keys_name)

        else:
            self.shm_keys_allocation.remap(size)

        self.shm_keys = np.ndarray(
            shape=shape,
            dtype=self.keys_data_type,
            buffer=self.shm_keys_allocation.buf
        )

        self.logger.info("Process %d extension successfully mapped." % self.pid)

        # Collapse the pile.
        self.collapse_pile(old_shape)

        self.logger.info("Process %d pile successfully collapsed." % self.pid)

    @log_timer
    def request_extension(self, size, create=False):
        self.con.send(
            (
                "shm_extend",
                (
                    self.shm_keys_name,
                    size,
                    create
                )
            )
        )
        result = self.con.recv()

        # Set shared_memory details.
        if create:
            self.shm_allocations["keys"]["shm_name"] = result
            self.shm_allocations["keys"]["active"] = True

        self.shm_keys_size = size

    @log_timer
    def collapse_pile(self, old_shape):
        size = self.pile_threshold + 1
        list_of_arrays = [arr for arr in self.yield_kmers()]

        # Check arrays are C contiguous.
        for arr in list_of_arrays:
            if not arr.flags["C_CONTIGUOUS"]:
                raise ValueError("Piled array is not C-contiguous!")

        filler(size, old_shape, list_of_arrays)

        # Reset the pile.
        self.pile = []

    def finish_import(self, *args):
        self.check_collapse()
        self.tell_size()

    def check_collapse(self):
        # Check if we need one final collapsing of the pile.
        if self.pile_length > 0:
            self.collapse()

        self.logger.info("Process %d finished." % self.pid)

    def tell_size(self):
        # Check for case where we have received no work.
        if self.shm_keys is not None and not self.is_supposed_to_die:
            self.con.send(
                (
                    "keys_size",
                    self.shape
                )
            )

            values_shm_name = self.con.recv()
            values_shape = self.shape[:1]

            # Initiate values allocation.
            self.shm_values_name = values_shm_name
            self.shm_values_size = self.product(values_shape, self.values_data_type_size)
            self.shm_values_allocation = ResizableSharedMemory(values_shm_name)
            self.shm_values = np.ndarray(
                shape=values_shape,
                dtype=self.values_data_type,
                buffer=self.shm_values_allocation.buf
            )

            self.shm_allocations["values"]["active"] = True

    def map(self, data):
        window, shm_data, node_id = data
        kmers = self.view_kmers(window, shm_data)

        self.logger.info("Process %d mapping %d." % (self.pid, node_id))
        binarysearch_put(
            kmers,
            self.shm_keys,
            self.shm_values,
            self.lca_matrix,
            node_id
        )

        # Demonstrate finished.
        kmers[:] = 0

    def save_shutdown(self, *args):
        self.save_data(*args)
        self.close_stores()
        self.close_shm()

    def save_data(self, *args):

        if self.shm_allocations["keys"]["active"] and self.shm_allocations["values"]["active"]:

            # Find our offset.
            self.con.send(
                (
                    "save",
                    self.shape
                )
            )

            arr_offset, db_path = self.con.recv()

            for arr in ("keys", "values"):
                self.shm_allocations[arr]["write_offset"] = arr_offset

            # Wait until it is my turn to open file.
            while self.loop_condition(self.lock_value.value != arr_offset):
                self.timeout()

            # Save data to disk.
            db = TablesDb(db_url=db_path, edit=True)

            for arr_name in ("keys", "values"):
                writer = db.write_keys if arr_name == "keys" else db.write_values
                arr = self.shm_allocations[arr_name]["array"]

                rows_written = writer(
                    arr=self.shm_allocations[arr_name]["array"],
                    offset=self.shm_allocations[arr_name]["write_offset"]
                )

                if rows_written != arr.shape[0]:
                    self.logger.info("Process %d writing of data failed!" % self.pid)
                else:
                    self.logger.info("Process %d successfully wrote data to %s!" % (self.pid, db_path))

            db.close()

            with self.lock_value.get_lock():
                self.lock_value.value += arr.shape[0]

        # Case where no data has been received.
        else:
            if not self.is_supposed_to_die:
                self.log("Warning! A child has gone this build without any work.")

                self.con.send(
                    (
                        "save",
                        None
                    )
                )

                receipt = self.con.recv()
                self.log(receipt)

    def close_stores(self):
        shm_checks = (
            (
                self.shm_allocations[col]["active"],
                self.shm_allocations[col]["allocation"]
            )
            for col in ("keys", "values")
        )

        for shm_check, shm_all in shm_checks:
            if shm_check:

                # Send an unlink and kill request to the main process.
                try:
                    shm_all.unlink()
                    shm_all.close()

                except FileNotFoundError:
                    pass

        # Log my death.
        self.logger.info("%d closed stores." % self.pid)

    def close_shm(self):
        for allocation in self.in_allocations.values():
            allocation.close()




import os
from typing import List, Set
from expam.cli.main import CommandGroup, ExpamOptions, clear_logs
from expam.database import FileLocationConfig
from expam.database.build import main as expam
from expam.database.config import ExpamDatabaseDoesNotExistError, JSONConfig, create_database, make_database_config, validate_database_file_configuration
from expam.logger import Timer
from expam.utils import die, ls, make_path_absolute


class BuildCommand(CommandGroup):
    commands: Set[str] = {
        'quickrun', 'default_db', 'create',
        'build', 'print', 'add', 'remove', 'set'
    }

    def __init__(
        self, config: FileLocationConfig,
        k: int, n: int, s: int, phylogeny_path: str, pile_size: int,
        files: List[str], group: str, first_n: int
    ) -> None:
        super().__init__()
        self.config: FileLocationConfig = config

        self.k = k
        self.n = n
        self.s = s
        self.phylogeny_path = phylogeny_path
        self.pile_size = pile_size

        self.files = files
        self.group = group
        self.first_n = first_n

    @classmethod
    def take_args(cls: CommandGroup, args: ExpamOptions) -> dict:
        k, n, s, pile_size, first_n = cls.parse_ints(args.k, args.n, args.s, args.pile, args.first_n)

        return {
            'config': make_database_config(args.db_name),
            'k': k,
            'n': n,
            's': s,
            'phylogeny_path': args.phylogeny,
            'pile_size': pile_size,
            'files': args.directory,
            'group': None if args.groups is None else args.groups[0][0],
            'first_n': first_n
        }

    def validate_database(self):
        try:
            validate_database_file_configuration(self.config)
        except ExpamDatabaseDoesNotExistError:
            die("%s is not a database." % self.config.database)

    """
    Quickrun command
    ================
    
    """
    def quickrun(self):
        self.create()
        self.set()
        self.add()
        self.build()

    """
    Default database command
    ========================
    
    """
    def default_db(self):
        self.validate_database()

        print("export EXPAM_DB_DIR=%s" % self.config.database)

    """
    Create database command
    =======================
    
    """
    def create(self):
        create_database(self.config)

    """
    Build database command
    ======================
    
    """
    def build(self):
        self.validate_database()
        conf = JSONConfig(self.config.conf)

        # Check if a phylogeny has been provided.
        if not self.check_phylogeny(conf):
            die("No phylogeny set for this database.")

        # Read configuration file.
        k, n, phylogeny, genome_paths, pile = conf.get_build_params()

        clear_logs(self.config.logs)

        # Set building database.
        try:
            with Timer() as t:
                expam(
                    db_path=self.config.base,
                    genome_paths=genome_paths,
                    phylogeny_path=phylogeny,
                    k=k,
                    n=n - 1,  # Account for main process.
                    pile_size=pile,
                )

            print("expam: " + str(t))

        finally:
            print(f'\nPID - {os.getpid()} dying...')

    @staticmethod
    def check_phylogeny(conf):
        if conf['phylogeny_path'] is None:
            return False
        else:
            return True

    """
    Print database command
    ======================
    
    """
    def print(self):
        self.validate_database()

        conf: JSONConfig = JSONConfig(self.config.conf)
        print(conf)

    """
    Add sequences to database command
    =================================
    
    """
    def add(self):
        self.validate_database()

        if self.files is None:
            die("No input directory provided!")
        
        for directory in self.files:
            self.add_sequences(directory)

    def add_sequences(self, path):
        self._modify_config(path, add=True)

    def _modify_config(self, directory, add=True):
        base_url = make_path_absolute(
            dir=directory,
            base_dir=os.getcwd()
        )

        # Get list of sequences in base_url.
        seq_names = ls(base_url)

        if (self.first_n is None) or (self.first_n > len(seq_names)):
            self.first_n = len(seq_names)

        seq_dirs = [
            os.path.join(base_url, seq_names[i])
            if seq_names[i] != base_url else seq_names[i]
            for i in range(self.first_n)
        ]

        # Append directory to database configuration.
        conf: JSONConfig = JSONConfig(self.config.conf)

        if add:
            conf.add_sequence(seq_dirs, group=self.group)
            print("Added %d files from %s." % (len(seq_dirs), base_url))
        else:
            conf.remove_sequence(seq_dirs, group=self.group)
            print("Removed %d files from %s." % (len(seq_dirs), base_url))

        conf.save()

    """
    Remove sequences to database command
    ====================================
    
    """
    def remove(self):
        self.validate_database()

        for directory in self.files:
            self.remove_sequences(directory)

    def remove_sequences(self, path):
        self._modify_config(path, add=False)

    
    """
    Set database settings
    =====================
    
    """
    def set(self):
        self.validate_database()

        conf = JSONConfig(url=self.config.conf)

        if self.group is None:
            conf.set(k=self.k, n=self.n, s=self.s, phylogeny_path=self.phylogeny_path, pile=self.pile_size)
        else:
            conf.group_set(group_name=self.group, k=self.k, s=self.s)

        conf.save()



#cython: infer_types=True

cimport cython
from libc.stdlib cimport malloc, free
import numpy as np
cimport numpy as np

ctypedef np.uint64_t uint64_t
ctypedef np.int64_t int64_t

# Custom C scripts.
cdef extern from "mfil.c":
    cdef int64_t multidisjoint(int n_arrays, int n_columns, uint64_t **arr_ptrs, int64_t *sizes)
    cdef void k_way_merge(uint64_t **arr_ptrs, int64_t *indices, uint64_t n, uint64_t k, uint64_t chunks)

def filler(uint64_t k, uint64_t old_shm_size, list list_of_arrays):
    """Merge-sort all arrays contained in list_of_arrays into one array,
    namely the first element in this list of arrays. This means that the first
    list of list_of_arrays needs to be large enough to fit the distinct values
    of all other elements, however it can itself contain values.

    :param old_shm_size: The number of values contained in the first array.
    :param list_of_arrays: List containing arrays to be merged.
    :return: None, in-place on first array.
    :rtype: None

    """
    cdef:
        uint64_t n, n_arrays, chunks
        uint64_t **arr_ptrs
        int64_t[::1] indices
        uint64_t[:, ::1] temp_arr
        int i

    n = list_of_arrays[0].shape[0]
    n_arrays = len(list_of_arrays)
    chunks = list_of_arrays[0].shape[1]

    # Pointers on each of the arrays.
    arr_ptrs = <uint64_t **>malloc(k * sizeof(uint64_t *))

    if not arr_ptrs:
        raise MemoryError

    # Pointer to current index value.
    indices = np.zeros(k, dtype=np.int64)

    # Type cast each of the input lists and make pointer.
    for i in range(k):
        if i < n_arrays:
            temp_arr = list_of_arrays[i]  # Cast array.
            arr_ptrs[i] = &temp_arr[0, 0]  # Declare pointer on array.

            # Set the current index (bottom -> up merging).
            if i == 0:
                indices[i] = old_shm_size - 1
            else:
                indices[i] = temp_arr.shape[0] - 1

        # Add dummy indices until we can make a binary tree.
        else:
            arr_ptrs[i] = &temp_arr[0, 0]
            indices[i] = -1

    try:
        k_way_merge(arr_ptrs, &indices[0], n, k, chunks)

    finally:
        free(arr_ptrs)

def disjoint(list list_of_arrays):
    """Iteratively remove elements from the first array until the values left
    are distinct, with respect to the values contained in the other arrays.

    :param list_of_arrays: List of sorted arrays.
    :return: Number of distinct elements left in first array.
    :rtype: int

    """
    cdef:
        int n, chunks, i
        int64_t final_size
        int64_t[::1] sizes
        uint64_t **arr_ptrs
        uint64_t[:, ::1] temp_arr

    n = len(list_of_arrays)
    chunks = list_of_arrays[0].shape[1]
    final_size = 0  # Set default.

    sizes = np.zeros(n, dtype=np.int64)
    arr_ptrs = <uint64_t **>malloc(n * sizeof(uint64_t))

    if not arr_ptrs:
        raise MemoryError

    # Establish pointer on each array.
    for i in range(n):
        temp_arr = list_of_arrays[i]    # Cast array.
        arr_ptrs[i] = &temp_arr[0, 0]   # Declare pointer on array.

        # Establish array size.
        sizes[i] = list_of_arrays[i].shape[0]

    try:
        final_size = multidisjoint(n, chunks, arr_ptrs, &sizes[0])

    finally:
        # Free heap arrays.
        free(arr_ptrs)

    return final_size



from collections import namedtuple
from functools import partial
from math import floor
from multiprocess import Pool   # Uses dill for serialising.
from multiprocessing import shared_memory
import os
import re
import shutil
from typing import List, Mapping, Tuple, Union

import numpy as np
import pandas as pd

from expam.classify import ResultsPathConfig
from expam.classify.config import create_tax_results, load_results_config
from expam.classify.process import ExpamClassifierProcesses
from expam.classify.taxonomy import TaxonomyNCBI
from expam.database import EXPAM_TEMP_EXT, TIMEOUT, UNION_RATIO, FileLocationConfig
from expam.database.config import load_database_config
from expam.database.db import SharedDB
from expam.process.classify import ClassifyWorker
from expam.process.manager import ControlCenter
from expam.process.reads import ReadWorker
from expam.tree.tree import Index
from expam.utils import ls, yield_csv


def run_classifier(
    read_paths: str,
    out_dir: str,
    db_dir: str,
    k: int,
    n: int,
    phylogeny_path: str,
    keys_shape: Tuple[int],
    values_shape: Tuple[int],
    logging_dir: str,
    taxonomy: bool = False, cutoff: float = 0.0, groups: Union[None, List[tuple]] = None,
    keep_zeros: bool = False, cpm: float = 0.0, use_node_names: bool = True,
    phyla: bool = False, name_taxa: Union[Mapping[str, str], None] = None,
    colour_list: List[str] = None, paired_end: bool = False, alpha: float = 1.0,
    log_scores: bool = False, itol_mode: bool = False,
    sep: str = ","
):
    output_config: ResultsPathConfig = load_results_config(out_dir, create=True)
    database_config: FileLocationConfig = load_database_config(db_dir)

    # Load the kmer dictionary.
    print("Loading the map and phylogeny.\n")
    db_kmers: SharedDB = SharedDB(db_dir, keys_shape, values_shape)

    try:
        # Get name mapping.
        index, phylogeny_index = name_to_id(phylogeny_path)

        # Load LCA Matrix.
        lca_matrix = np.load(database_config.lca_matrix)

        # Initialise the distribution.
        d = Distribution(
            k=k,
            kmer_db=db_kmers,
            index=phylogeny_index,
            lca_matrix=lca_matrix,
            read_paths=read_paths,
            out_dir=out_dir,
            logging_dir=logging_dir,
            temp_dir=output_config.temp,
            paired_end=paired_end,
            alpha=alpha,
            sep=sep,
            n_processes=n
        )

        # Run the classification with n processes.
        d.run()

        # Load accession ids.
        sequence_ids = list(yield_csv(database_config.accession_id))

        # Update accession ids in map.
        for (fname, accn_id, tax_id) in sequence_ids:
            if accn_id != "None":
                phylogeny_index[fname].accession_id = accn_id

            if tax_id != "None":
                phylogeny_index[fname].taxid = tax_id

        # Output results.
        results = ClassificationResults(
            index=index,
            phylogeny_index=phylogeny_index,
            results_config=output_config,
            groups=groups,
            keep_zeros=keep_zeros,
            cutoff=cutoff,
            cpm=cpm,
            use_node_names=use_node_names,
            phyla=phyla,
            name_taxa=name_taxa,
            colour_list=colour_list,
            log_scores=log_scores,
            sep=sep
        )

        if taxonomy:
            # Attempt to update taxon ids.
            tax_obj: TaxonomyNCBI = TaxonomyNCBI(database_config)
            tax_obj.accession_to_taxonomy()

            name_to_lineage, taxon_to_rank = tax_obj.load_taxonomy_map()
            results.to_taxonomy(name_to_lineage, taxon_to_rank)

        results.draw_results(itol_mode=itol_mode)
    finally:
        # Close shared memory.
        keys_shm = shared_memory.SharedMemory(
            name=db_kmers.keys_shm_name,
            create=False
        )
        values_shm = shared_memory.SharedMemory(
            name=db_kmers.values_shm_name,
            create=False
        )
        keys_shm.unlink()
        keys_shm.close()
        values_shm.unlink()
        values_shm.close()

        shutil.rmtree(output_config.temp)


def name_to_id(phylogeny_path: str):
    # Load Index from Newick file.
    with open(phylogeny_path, "r") as f:
        newick_str = f.read().strip()

    _, index_map = Index.from_newick(newick_str)

    index = {}
    for i, loc in enumerate(index_map.pool):
        index[loc.name] = i

    return index, index_map


ResultsColumns = namedtuple("ResultsColumns", ["c_perc", "c_cumul", "c_count", "s_perc", "s_cumul", "s_count"])

class Distribution:
    _cols = ResultsColumns("Cumulative Classified Percentage", "Cumulative Classified Count", "Raw Classified Count",
                           "Cumulative Split Percentage", "Cumulative Split Count", "Raw Split Count")

    def __init__(self, k, kmer_db, index: Index, lca_matrix, read_paths, out_dir, temp_dir, logging_dir, alpha,
                 keep_zeros=False, cutoff=0.0, cpm=0.0, paired_end=False, sep=",", n_processes: int = 1):
        """

        :param k: Int
        :param kmer_db: SharedDB
        :param lca_matrix: np.ndarray
        :param url: str
        :param out_dir: str
        """
        # Set up processing references.
        self.k = k
        self.alpha = alpha
        self.kmer_db = kmer_db
        self.lca_matrix = lca_matrix
        self.out_dir = out_dir
        self.temp_dir = temp_dir
        self.logging_dir = logging_dir

        self.index: Index = index
        self.node_names = [node.name if i > 0 else "unclassified" for i, node in enumerate(index.pool)]

        self.keep_zeros = keep_zeros
        self.cutoff = cutoff
        self.cpm = cpm

        # Url containing reads.
        self.results_config: ResultsPathConfig = load_results_config(out_dir, create=True)
        self.read_paths = read_paths
        self.paired_end = paired_end

        self.SEP = sep
        self.n_processes = n_processes

    def run(self):
        """
        Begin processing reads in self.url.
        We have some collection of processes accessing reads from storage
        and then another collection of processes that analyse the reads.

        """
        file_queue = self.prepare_queue(self.paired_end)
        n_classifiers = floor(UNION_RATIO * self.n_processes)
        n_readers = self.n_processes - n_classifiers

        mp_config = {
            "name": "expam",
            "phases": ["classify"],
            "layers": {
                "class": ReadWorker,
                "class_args": {
                    "paired_mode": self.paired_end,
                    "n_classifiers": n_classifiers,
                    "logging_dir": self.logging_dir,
                },
                "parent": ControlCenter,
                "child": {
                    "class": ClassifyWorker,
                    "class_args": {
                        "k": self.k,
                        "kmer_db": self.kmer_db,
                        "lca_matrix": self.lca_matrix,
                        "out_dir": self.results_config.phy,
                        "temp_dir": self.temp_dir,
                        "n_readers": n_readers,
                        "logging_dir": self.logging_dir,
                        "alpha": self.alpha
                    },
                    "parent": ReadWorker,
                    "child": None,
                    "n": n_classifiers
                },
                "n": n_readers
            },
            "timeout": TIMEOUT
        }

        process_network = ExpamClassifierProcesses.from_method_dict(self.logging_dir, mp_config)

        # Load jobs.
        for file_dir in file_queue:
            process_network.load_job(phase="classify", job=file_dir)

        process_network.run()

        # Process and combine outputs.
        output_files = self.combine_raw_read_output()
        self.create_sample_summaries(output_files)

    def combine_raw_read_output(self):
        temporary_files = ls(self.temp_dir, ext=".reads_%s" % EXPAM_TEMP_EXT)
        base_file_names = {
            re.match(r'(\S+)_\d+.reads_%s' % EXPAM_TEMP_EXT, os.path.basename(file_name)).group(1)
            for file_name in temporary_files
        }

        result_files = []
        for sample_name in base_file_names:
            results_file_name = os.path.join(self.results_config.phy_raw, sample_name + ".csv")
            raw_files = [file for file in temporary_files if re.match(r'.*{base}_\d+.reads_{ext}$'.format(base=sample_name, ext=EXPAM_TEMP_EXT), file)] 
            
            self.concatenate_files(results_file_name, *raw_files)
            result_files.append((results_file_name, sample_name))

        return result_files

    def create_sample_summaries(self, result_files):
        nprocs = min(self.n_processes, len(result_files))
        ###
        ### Import raw counts into DataFrame.
        def _raw_sample_to_series(results_file, sample_name, node_names):
            sl = pd.Series(0, index=node_names, name=sample_name)
            ml = pd.Series(0, index=node_names, name=sample_name)
            with open(results_file, 'r') as f:
                for line in f:
                    parts = line.split("\t")
                    idx = int(parts[2].lstrip("p"))
                    if parts[0] == "S":
                        ml.iloc[idx] += 1
                    else:
                        sl.iloc[idx] += 1
            return sample_name, sl, ml

        ###
        with Pool(processes=nprocs) as mp_pool:
            raw_sample_series = sorted(
                mp_pool.starmap_async(
                    partial(_raw_sample_to_series, node_names=self.node_names),
                    result_files).get(),
                key = lambda t: t[0])   # Sorted in order of sample names.
        
        ###
        ###

        ###
        ### Create sample-wise summaries.
        def remove_zeros(*dfs):
            mask = None
            for df in dfs:
                non_zero = df.apply(lambda r: ((r > 0).any()) | (r.name == 'unclassified'), axis=1)
                mask = non_zero if mask is None else mask | non_zero
            for df in dfs:
                yield df.loc[mask, :]

        ###
        def _sample_summary(
            sample_name: str, sl: pd.Series, ml: pd.Series, out_base: str,
            node_names: List[str], cols: ResultsColumns, index: Index,
            cutoff: int, cpm: float, sep: str):
            df = pd.DataFrame(0, index=node_names, columns=tuple(cols))
            df.loc[:, cols.c_cumul] = sl.loc[:]
            df.loc[:, cols.c_count] = sl.loc[:]
            df.loc[:, cols.s_cumul] = ml.loc[:]
            df.loc[:, cols.s_count] = ml.loc[:]

            # Make counts cumulative.
            cumulative_columns = [cols.c_cumul, cols.s_cumul]
            for node in index.pool[1:][::-1]:
                df.loc[node.name, cumulative_columns] += df.loc[node.children, cumulative_columns].sum(axis=0)

            # Employ cutoff.
            total_reads = df.loc[:, cols.c_count].sum() + df.loc[:, cols.s_count].sum()
            if cutoff > 0 or cpm > 0.0:
                cutoff = max((total_reads / 1e6) * cpm, cutoff)
                mask = (df.loc[:, cols.c_cumul] < cutoff) & (df.loc[:, cols.s_cumul])
                mask[0] = True  # Keep unclassified counts.
                df.loc[mask, :] = 0

            # Check for removal of zero rows.
            if not self.keep_zeros:
                (df, ) = remove_zeros(df)

            # Insert percentage form.
            df.loc[:, cols.c_perc] = round(100 * df.loc[:, cols.c_cumul] / total_reads, 3).map(str) + "%"
            df.loc[:, cols.s_perc] = round(100 * df.loc[:, cols.s_cumul] / total_reads, 3).map(str) + "%"

            # Write to disk.
            out_file = os.path.join(out_base, sample_name + ".csv")
            fixer = lambda name: 'p' + name if (name != "unclassified") and (index[name].type == "Branch") else name
            df.index = [fixer(n) for n in df.index]
            df.to_csv(out_file, sep=sep)

        ###
        ###

        with Pool(processes=nprocs) as mp_pool:
            f = partial(
                _sample_summary,
                out_base=self.results_config.phy,
                node_names=self.node_names, cols=self._cols, index=self.index,
                cutoff=self.cutoff, cpm=self.cpm, sep=self.SEP
            )
            mp_pool.starmap_async(f, raw_sample_series).get()


        ###
        ### Save summarised results to disk.
        (_, sl_series_list, ml_series_list) = zip(*raw_sample_series)
        sldf = pd.concat(sl_series_list, axis=1)
        mldf = pd.concat(ml_series_list, axis=1)

        # Remove null rows.
        if not self.keep_zeros:
            (sldf, mldf) = remove_zeros(sldf, mldf)

        # Drop unclassified from split DataFrame.
        try:
            mldf.drop(index="unclassified", inplace=True)
        except KeyError:    # If no unclassified, this will have already been removed.
            pass

        # Save counts to disk.
        self.save_df(sldf, self.results_config.phy_classified)
        self.save_df(mldf, self.results_config.phy_split)

    def save_df(self, df: pd.DataFrame, out_file: str):
        self.fix_df_index(df)
        df.to_csv(out_file, sep=self.SEP)

    def fix_df_index(self, df: pd.DataFrame):
        fixer = lambda name: 'p' + name if (name != "unclassified") and (self.index[name].type == "Branch") else name
        df.index = [fixer(n) for n in df.index]

    @staticmethod
    def concatenate_files(dest, *files):
        with open(dest, 'w') as out_f:
            for file in files:
                # Write data to concatenated file.
                with open(file, 'r') as in_f:
                    out_f.write(in_f.read().strip() + "\n")

                # Remove temporary file.
                os.remove(file)

    @staticmethod
    def best_match(file_name, file_list):
        def string_intersect(str_one, str_two):
            l = min(len(str_one), len(str_two))

            for i in range(l):
                if str_one[i] != str_two[i]:
                    return str_one[:i]
            else:
                return str_one[:l]

        match_ind = 0
        best_match = string_intersect(file_name, file_list[0])

        for i in range(1, len(file_list)):
            current_match = string_intersect(file_name, file_list[i])

            if len(current_match) > len(best_match):
                best_match = current_match
                match_ind = i

        return match_ind

    def prepare_queue(self, paired_end=False):
        file_queue = []

        for path in self.read_paths:
            # Create a list of all the files containing reads to be processed.
            file_list = [
                file_name
                for file_name in ls(path)
                if re.match(r"(\S+)\.(?:fa|fna|ffn|fq|fasta|fastq)(?:\.tar)*(?:\.gz)*$", file_name)
            ]

            if paired_end:  # This implies path is a folder, not a file.
                while file_list:
                    next_file = file_list.pop()

                    # Find buddy.
                    match_ind = self.best_match(next_file, file_list)
                    paired_file = file_list.pop(match_ind)

                    file_queue.append((next_file, paired_file))
            else:
                for file_name in file_list:
                    file_queue.append((file_name, ))

        return file_queue

    @staticmethod
    def _initialise_results(names, unclassified=True):
        """
        Results are returned in spectral form:
            coord --> # hits at coord.

        :param names: List - node names in the tree.
        :return: pd.Series()

        """
        index = (unclassified * ["unclassified"]) + names
        return pd.Series(
            data=np.zeros(len(names) + unclassified, dtype=int),
            index=index
        )


class ClassificationResults:
    _c_perc = "Cumulative Classified Percentage"
    _c_cumul = "Cumulative Classified Count"
    _c_count = "Raw Classified Count"
    _s_perc = "Cumulative Split Percentage"
    _s_cumul = "Cumulative Split Count"
    _s_count = "Raw Split Count"
    _rank = "Rank"
    _scientific_name = "Scientific Name"

    def __init__(self, index, phylogeny_index, results_config, groups=None, keep_zeros=False, cutoff=0.0, cpm=0.0,
                 use_node_names=True, phyla=False, name_taxa=None, colour_list=None, circle_scale=1.0,
                 log_scores=False, sep=","):
        self.index = index
        self.phylogeny_index: Index = phylogeny_index
        self.results_config: ResultsPathConfig = results_config

        self.groups = groups  # [(colour, (name1, name2, ...)), ...]
        self.keep_zeros = keep_zeros

        self.cutoff = cutoff  # Cutoff by counts.
        self.cpm = cpm  # Cutoff by counts per million.
        self.use_node_names = use_node_names

        self.phyla = phyla
        self.colour_list = colour_list
        self.name_taxa = name_taxa
        self.circle_scale = circle_scale
        self.log_scores = log_scores

        if self.phyla and self.name_taxa is None:
            raise Exception("Inclusion of phyla requires mapping of names to taxonomic lineages!")

        self.taxid_immediate_children = {"1": set()}  # Map from tax_id -> immediate children.
        self.tax_id_pool = ["1"]  # Children must appear later than parent this list.

        self.SEP = sep

    def to_taxonomy(self, name_to_lineage, taxon_to_rank):
        col_names = [self._c_perc, self._c_cumul, self._c_count, self._s_perc, self._s_cumul, self._s_count, self._rank, self._scientific_name]

        class_counts = pd.read_csv(self.results_config.phy_classified, sep=self.SEP, index_col=0, header=0)
        split_counts = pd.read_csv(self.results_config.phy_split, sep=self.SEP, index_col=0, header=0)

        # Get rid of phylogenetically printed node names.
        def fix_index(df: pd.DataFrame):
            df.index = [
                index.lstrip("p")
                if index not in self.phylogeny_index._pointers
                else index
                for index in df.index
            ]

        fix_index(class_counts)
        fix_index(split_counts)

        samples = class_counts.columns.tolist()
        phylogeny_nodes = split_counts.index.tolist()

        # Map phylogeny index to taxonomy.
        phylogeny_to_taxonomy, tax_id_data = {}, {}
        for phylogeny_node in phylogeny_nodes:
            tax_id, tax_id_map = self.map_phylogeny_node(phylogeny_node, name_to_lineage, taxon_to_rank)

            phylogeny_to_taxonomy[phylogeny_node] = tax_id
            tax_id_data.update(tax_id_map)

        # Get relevant taxon ids.
        tax_ids, ranks, lineages = ["1"], ["root"], [""]
        for tax_id in self.tax_id_pool[1:]:
            rank, lineage = tax_id_data[tax_id]

            tax_ids.append(tax_id)
            ranks.append(rank)
            lineages.append(lineage)

        # Make taxonomic output folder structure.
        create_tax_results(self.results_config)

        for sample_name in samples:
            #
            # Map summary output to taxonomy.
            #

            df = pd.DataFrame(0, index=["unclassified"] + tax_ids, columns=col_names)
            total_counts = class_counts[sample_name].sum() + split_counts[sample_name].sum()

            df.loc[tax_ids[0]:, self._rank] = ranks
            df.loc[tax_ids[0]:, self._scientific_name] = lineages

            # Manually insert unclassified counts.
            df.loc["unclassified", self._c_count] = class_counts.loc["unclassified", sample_name]
            df.loc["unclassified", self._c_cumul] = df.loc["unclassified", self._c_count]
            df.loc["unclassified", self._c_perc] = round(100 * df.loc["unclassified", self._c_cumul] / total_counts, 3)

            for col, counts_df in ((self._c_count, class_counts), (self._s_count, split_counts)):
                for node in counts_df.index[1:]:
                    df.loc[phylogeny_to_taxonomy[node], col] += counts_df.loc[node, sample_name]

            # Make counts cumulative.
            # We may not be able to use groupby, but we can iterate through the taxid
            # pool once (in hierarchical order).
            for tax_id in self.tax_id_pool[::-1]:
                for count_col, cumul_col in (
                    (self._c_count, self._c_cumul),
                    (self._s_count, self._s_cumul)
                ):
                    df.loc[tax_id, cumul_col] = df.loc[tax_id, count_col] + df.loc[list(self.taxid_immediate_children[tax_id]), cumul_col].sum(axis=0)

            # Format percentage signs.
            df.loc[:, self._c_perc] = round(100 * df.loc[:, self._c_cumul] / total_counts, 3).map(str) + "%"
            df.loc[:, self._s_perc] = round(100 * df.loc[:, self._s_cumul] / total_counts, 3).map(str) + "%"

            # Employ cutoff.
            cutoff = max(self.cutoff, (total_counts / 1e6) * self.cpm)
            df = df[(df[self._c_cumul] > cutoff) | (df[self._s_cumul] > cutoff) | (df.index == 'unclassified')]
            df.to_csv(os.path.join(self.results_config.tax, sample_name + ".csv"), sep=self.SEP, header=True)

            #
            # Map raw read output to taxonomy.
            #
            raw_counts_file = os.path.join(self.results_config.phy_raw, sample_name + ".csv")
            output_file = os.path.join(self.results_config.tax_raw, sample_name + '.csv')

            with open(raw_counts_file, 'r') as in_f, open(output_file, 'w') as out_f:
                for line in in_f:
                    read_data = line.strip().split("\t")[:-1]

                    if read_data[0] == 'U':
                        read_data[2] = 'NA'
                    else:
                        phy_id = int(read_data[2].lstrip('p'))
                        phylogeny_node = self.phylogeny_index.pool[phy_id].name
                        read_data[2] = phylogeny_to_taxonomy[phylogeny_node]

                    out_f.write("\t".join(read_data) + "\n")

    def map_phylogeny_node(self, node_name, name_to_lineage, taxon_to_rank):
        """
        Create a map from (node_name) --> (taxid, rank, taxonomic lineage).
        
        :param node_name: Name of clade.
        :param name_to_lineage: Dict[Accession ID, Tuple[Taxonomic Lineage]]
        :param taxon_to_rank: Dict[Taxonomic Name, Tuple[TaxID, Rank]]
        """
        leaves = self.get_leaves(node_name)
        lineages = [name_to_lineage[leaf] for leaf in leaves]
        common_lineage = self.tuple_intersect(*lineages)

        if not common_lineage:
            phy_tax_id, phy_rank = "1", "root"
        else:
            phy_tax_id, phy_rank = taxon_to_rank[common_lineage[-1]]

        taxid_rank_and_lineage = {}
        pool_index = 1
        for i in range(len(common_lineage)):
            # Store taxonomic data.
            parent_tax_id, parent_rank = taxon_to_rank[common_lineage[i]]
            taxid_rank_and_lineage[parent_tax_id] = [parent_rank, " ".join(common_lineage[:i + 1])]
            
            # Add top node as child of root.
            if i == 0:
                self.taxid_immediate_children["1"].add(parent_tax_id)

            # Set tax_id in pool - a list of taxids such that child taxids are necessarily further
            # in the last than any ancestors.
            try:
                pool_index = self.tax_id_pool.index(parent_tax_id) + 1
            except ValueError:
                self.tax_id_pool.insert(pool_index, parent_tax_id)
                pool_index += 1

            # Store immediate child of taxid.
            if i + 1 < len(common_lineage):
                child_tax_id, child_rank = taxon_to_rank[common_lineage[i + 1]]

                try:
                    self.taxid_immediate_children[parent_tax_id].add(child_tax_id)
                except KeyError:
                    self.taxid_immediate_children[parent_tax_id] = {child_tax_id}

            elif parent_tax_id not in self.taxid_immediate_children:
                self.taxid_immediate_children[parent_tax_id] = set()

        return phy_tax_id, taxid_rank_and_lineage

    def get_leaves(self, node_name):
        return self.phylogeny_index.get_child_leaves(node_name)

    @staticmethod
    def tuple_intersect(*tuples):
        intersection = ()

        length = min([len(tup) for tup in tuples])
        for i in range(length):
            item = tuples[0][i]

            for j in range(1, len(tuples)):
                if tuples[j][i] != item:
                    return intersection

            else:
                intersection += (item,)

        return intersection

    def draw_results(self, itol_mode=False, flat_colour=False):
        # Draw classified tree.
        self.phylogeny_index.draw_results(
            self.results_config.phy_classified,
            os.path.join(self.results_config.base, "phylotree_classified.pdf"),
            skiprows=[1],
            groups=self.groups,
            cutoff=self.cutoff,
            cpm=self.cpm,
            colour_list=self.colour_list,
            name_to_taxon=self.name_taxa,
            use_phyla=self.phyla,
            keep_zeros=self.keep_zeros,
            use_node_names=self.use_node_names,
            log_scores=self.log_scores,
            itol_mode=itol_mode,
            flat_colour=flat_colour,
            sep=self.SEP
        )

        # Draw unclassified tree.
        self.phylogeny_index.draw_results(
            self.results_config.phy_split,
            os.path.join(self.results_config.base, "phylotree_splits.pdf"),
            groups=self.groups,
            cutoff=self.cutoff,
            cpm=self.cpm,
            colour_list=self.colour_list,
            name_to_taxon=self.name_taxa,
            use_phyla=self.phyla,
            keep_zeros=self.keep_zeros,
            use_node_names=self.use_node_names,
            log_scores=self.log_scores,
            itol_mode=itol_mode,
            flat_colour=flat_colour,
            sep=self.SEP
        )



from ctypes import c_uint64
import math
from multiprocessing import Pipe, Value, shared_memory
import os
import subprocess
from typing import List, Mapping
import numpy as np
from expam.database import CHUNK_SIZE, TIMEOUT, UNION_RATIO, FileLocationConfig, expam_dtypes
from expam.database.config import load_database_config
from expam.process.genome import ExtractWorker
from expam.process.manager import ControlCenter, ExpamProcesses
from expam.process.piler import UnionWorker
from expam.sequences import COMP_PARSE, check_suffix
from expam.tree.tree import Index, propose_lca
from expam.utils import ls


def main(
    db_path: str, genome_paths: List[str], phylogeny_path: str, k: int,
    n=None, n_extract=None, n_union=None, pile_size=None,
):
    # Configure number of processes.
    if n_union is not None and n_extract is not None:
        n_union, n_extract = int(n_union), int(n_extract)
    elif n is not None:
        n_union, n_extract = distribute_processes(n)
    else:
        raise ValueError("Number of processes is can't be defined!")

    # Processing parameter details.
    num_cols = calculate_n_chunks(k)
    ranges = make_ranges(n_union, k)
    genome_paths, maxsize = sort_by_size(genome_paths)

    # Create shm_allocations for in-place kmer processing.
    shm_allocations = prepare_kmer_allocations(maxsize, num_cols, n_extract)

    # Import phylogeny.
    _, index = import_phylogeny(phylogeny_path)
    # Create LCA matrix.
    node_to_index = {node.name: i for i, node in enumerate(index.pool) if i > 0}
    lca_matrix = make_lca_matrix(index, node_to_index)

    # Prepare pipes between main and union workers.
    main_cons, union_cons = make_pipes(n_union)
    # Lock system to prevent concurrent access to HDF5 file.
    lock_value = Value(c_uint64, lock=True)

    # Multiprocessing configuration.
    config: FileLocationConfig = load_database_config(db_path)
    mp_config = {
        "name": "expam",    # Job system title.
        "phases": [         # Two-pass algorithm.
            "import",       # Import sequences and take union.
            "map"           # Map kmer to LCA.
        ],
        "layers": {         # Top layer of child processes - extract sequence & kmers.
            "class": ExtractWorker,
            "class_args": { # Positional arguments to init class.
                "k": k,
                "shm_kmer_params": shm_allocations,
                "kmer_ranges": ranges,
                "node_map": node_to_index,
                "out_dir": db_path,
                "logging_dir": config.logs,
            },
            "parent": ControlCenter,                        # Receive work from main process.
            "child": {
                "class": UnionWorker,       # Take union of sets of kmers.
                "class_args": { # Each either length n_union or 1.
                    "main_con": union_cons,
                    "lca_matrix": lca_matrix,
                    "pile_size": pile_size,
                    "logging_dir": config.logs,
                    "lock_value": lock_value,
                },
                "parent": ExtractWorker,    # Receive work from ExtractWorkers.
                "child": None,  # Bottom of process hierarchy.
                "n": n_union,   # Number of processes.
            },
            "n": n_extract,
        },
        "timeout": TIMEOUT,     # Time to block for when checking queues.
    }

    process_network = ExpamProcesses.from_method_dict(main_cons, db_path, config.logs, mp_config)

    # Load files for importing stage.
    for file_dir in genome_paths:
        process_network.load_job(
            phase="import",
            job=file_dir
        )

    process_network.run()

    # Save LCA matrix.
    np.save(config.lca_matrix, lca_matrix)

    # Concatenate accession ID files.
    concatenate_accession_ids(config.phylogeny)

def distribute_processes(n):
    n_union = max(math.floor(UNION_RATIO * n), 1)
    n_extract = max(n - n_union, 1)

    return n_union, n_extract

def make_ranges(n, k):
    # Calculate hash offset for the k value.
    maxvalue = maximum_hash(k)

    # Interpolate between offset and maxvalue.
    dx = maxvalue // n
    return {
        i: (i * dx, (i + 1) * dx) if i < n - 1
        else (i * dx, maxvalue)
        for i in range(n)
    }

def import_phylogeny(phylogeny_path):
    print("Importing phylogeny...")

    # Load the phylogeny into an Index.
    genome_names, index = Index.load_newick(phylogeny_path)
    return genome_names, index

def maximum_hash(k):
    if k < CHUNK_SIZE:
        return 4 ** k - 1
    else:
        return 4 ** CHUNK_SIZE - 1

def sort_by_size(dirs):
    _file_suffixes = COMP_PARSE
    _suffix_check = check_suffix

    def _gzip_size(file_dir):
        file_stats = subprocess.run(["gzip", "-l", file_dir], capture_output=True)
        file_meta = [entry for entry in file_stats.stdout.split(b' ') if entry]
        return file_meta[5]

    def _fna_size(file_dir):
        return os.stat(file_dir).st_size

    def _get_file_size(file_dir):
        # Check if file is symlink.
        if os.path.islink(file_dir):
            file_dir = os.path.realpath(file_dir)

        for suffix in _file_suffixes:
            if _suffix_check(file_dir, suffix):
                return _gzip_size(file_dir)

        return _fna_size(file_dir)

    max_filename = max([len(f) for f in dirs])
    file_dt = np.dtype([
        ("filename", "S%d" % max_filename),
        ("filesize", np.int64)
    ])

    # Turn into structured array.
    file_array = np.array([
        (file_dir, _get_file_size(file_dir))
        for file_dir in dirs
    ], dtype=file_dt)

    # Sort by file size.
    file_array.sort(order="filesize")

    max_size = file_array["filesize"][-1]
    ordered_files = [
        filename.decode("utf8")
        for filename in file_array["filename"].tolist()
    ]

    # Assuming each char in the file is a base.
    return ordered_files, max_size

def prepare_kmer_allocations(rows, cols, n_processes):
    allocation_params = tuple()
    allocation_size = rows * cols * expam_dtypes.keys_dtype_size
    allocation_shape = (rows, cols)

    for _ in range(n_processes):
        # Create and 0 allocation.
        next_shm_allocation = shared_memory.SharedMemory(
            create=True,
            size=allocation_size
        )

        next_arr = np.ndarray(
            shape=allocation_shape,
            dtype=expam_dtypes.keys_dtype,
            buffer=next_shm_allocation.buf
        )
        next_arr[:] = 0

        # Pass on allocation details to children.
        allocation_params += ((next_shm_allocation.name, allocation_shape), )
        next_shm_allocation.close()

    return allocation_params

def calculate_n_chunks(k: int):
    return k // CHUNK_SIZE + (k % CHUNK_SIZE != 0)

def yield_coordinates(index: Index):
    n = len(index.pool)

    for i in range(1, n):
        for j in range(1, i):
            yield i, j, index.pool[i].coordinate, index.pool[j].coordinate

def compute_lca(i, j, coord_one, coord_two):
    return i, j, propose_lca(coord_one, coord_two)

def make_lca_matrix(index: Index, node_to_index: Mapping[str, str]):
    print("Creating LCA matrix...")

    def get_children(fixed_node, flexible_index_list):
        nonlocal index, matrix, node_to_index

        i = 0
        while i < len(flexible_index_list):
            for child in index.pool[flexible_index_list[i]].children:
                m = min(fixed_node, node_to_index[child])
                n = max(fixed_node, node_to_index[child])

                if matrix[n, m] == 0 and n != m:
                    flexible_index_list.append(node_to_index[child])

            i += 1

    n = len(index.pool)
    matrix = np.zeros((n, n), dtype=np.uint16)

    for i in range(1, n):
        subject = index.pool[i].coordinate
        subject_name = index.pool[i].name

        for j in range(1, i):
            if matrix[i, j] > 0:
                continue

            target = index.pool[j].coordinate
            target_name = index.pool[j].name

            lca_coordinate = propose_lca(subject, target)
            lca = index.coord(coordinate=lca_coordinate).name

            left_group = [i]
            right_group = [j]

            if lca not in [subject_name, target_name]:
                # Subject and target are on separate lineages.
                get_children(j, left_group)
                get_children(i, right_group)

            elif lca == subject_name:
                # Target is contained in the clade rooted at subject.
                get_children(i, right_group)

            elif lca == target_name:
                # Subject is contained in the clade rooted at target.
                get_children(j, left_group)

            for p in left_group:
                for q in right_group:
                    m = min(p, q)
                    n = max(p, q)

                    if n != m and matrix[n, m] == 0:
                        matrix[n, m] = node_to_index[lca]

    return matrix

def make_pipes(n):
    parent_cons, child_cons = [], []

    i = 0
    while i < n:
        parent_con, child_con = Pipe()

        parent_cons.append(parent_con)
        child_cons.append(child_con)

        i += 1

    return tuple(parent_cons), tuple(child_cons)

def concatenate_accession_ids(f_url):
    def get_data(file_url):
        with open(file_url, "r") as f:
            data = f.read()

        # Delete temporary data file.
        os.remove(file_url)

        return data

    def get_files(folder_url):
        return [
            file_url
            for file_url in ls(folder_url)
            if "accession_ids_" in file_url
        ]

    id_data = [
        get_data(os.path.join(f_url, file_name))
        for file_name in get_files(f_url)
    ]

    with open(os.path.join(f_url, "accession_ids.csv"), "w") as f:
        f.write("\n".join(id_data))






from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from collections import namedtuple
import datetime
import os
import shutil
from typing import Set

import matplotlib.pyplot as plt
import numpy as np

from expam import __version__
from expam.utils import die, ls, make_path_absolute, parse_float, parse_int

ExpamOptions = namedtuple(
    'ExpamOptions',
    [
        # Runtime arguments
        'command', 'db_name', 'k', 'n', 's', 'phylogeny', 'alpha',
        # Directory arguments
        'directory', 'out_url', 'truth_dir',
        # Parameter arguments
        'length', 'pile', 'error_rate', 'first_n', 'paired_end',
        # Summary arguments
        'plot', 'cutoff', 'cpm', 'taxonomy',
        # Plot arguments
        'groups', 'phyla', 'keep_zeros', 'ignore_names', 'colour_list', 'rank', 'log_scores', 'itol_mode', 'flat_colour',
        # Tree arguments
        'use_sourmash', 'use_rapidnj', 'use_quicktree'
    ]
)

def retrieve_arguments() -> ExpamOptions:
    parser = ArgumentParser(description="  expam CLI\n--------------\n", formatter_class=RawTextHelpFormatter)

    parser.add_argument('--version', action='version', version='%(prog)s ' + ".".join(str(v) for v in __version__))
    parser.add_argument("command", default=None,
                        help='\nCommand to execute. Valid commands include:\n'
                             '-------------------------------------------\n'
                             'create:-\tInitialise database.\n'
                             'build:-\t\tStart building database.\n'
                             'print:-\t\tPrint current database parameters.\n'
                             'run:-\t\tRun reads against database.\n'
                             'add:-\t\tAdd sequence to the database.\n'
                             'remove:-\tRemove sequence from database (only impacts future db builds).\n'
                             'set:-\t\tSet database build parameters.\n'
                             'to_taxonomy:-\t\tConvert results to taxonomic setting.\n'
                             'phylotree:-\t\tDraw results on phylotree.\n'
                             'draw_tree:-\t\tDraw the reference tree.\n'
                             'download_taxonomy:-\t\tDownload taxonomic information for reference seqeunces.\n'
                             'cutoff:-\t\tApply cutoff to some set of already processed classifications. THIS WILL OVERWRITE OLD RESULTS!\n'
                             'mashtree:-\tCreate mashtree from current sequences and add to database.\n'
                             'quickrun:-\tInitialise, set parameters and start building db (assumes\n'
                             '\t\t\tsequences all lie in the same folder).\n'
                             'make_reads:-\tUniformly sample reads of length l from some input sequence.\n'
                             '\t\tThis is for testing purposes only, and is not a replacement\n'
                             '\t\tfor actual read generating software.\n',
                        metavar="[command]")
    parser.add_argument("-db", "--db_name", dest="db_name",
                        help="Name of database.",
                        metavar="[database name]")
    parser.add_argument("-k", "--kmer", dest="k",
                        help="Length of mer used for analysis.",
                        metavar="[k value (int)]")
    parser.add_argument("-n", "--n-processes", dest="n",
                        help="Number of CPUs to use for processing.",
                        metavar="[n (int)]")
    parser.add_argument("-s", "--sketch", dest="s",
                        help="Sketch size for mash.",
                        metavar="[sketch size (int)]")
    parser.add_argument("-p", "--phylogeny", dest="phylogeny",
                        help="URL of Newick file containing phylogeny.",
                        metavar="[phylogeny URL]")
    parser.add_argument("-d", "--directory", dest="directory", action="append",
                        help="File URL, context depending on command supplied.",
                        metavar="[directory]")
    parser.add_argument("-l", "--length", dest="length",
                        help="Length of simulated reads.",
                        metavar="[read length]")
    parser.add_argument("-o", "--out", dest="out_url",
                        help="Where to save classification results.",
                        metavar="[out URL]")
    parser.add_argument("-y", "--pile", dest="pile",
                        help="Number of genomes to pile at a time (or inf).",
                        metavar="[pile size]")
    parser.add_argument("-e", "--error-rate", dest="error_rate",
                        help="Generate error in reads (error ~ reads with errors / reads).",
                        metavar="[error rate]")
    parser.add_argument("-t", "--truth", dest="truth_dir",
                        help="Location of truth dataset.")
    parser.add_argument("--plot", dest="plot", default=False, action="store_true",
                        help="Plot timing data of database build.")
    parser.add_argument("--first", dest="first_n", default=None,
                        help="Add first n genomes in folder.")
    parser.add_argument("--cutoff", dest="cutoff", default=0,
                        help="Ignore organisms with less than `cutoff` reads in results.")
    parser.add_argument("--cpm", dest="cpm", default=100,
                        help="Counts/million cutoff for read-count to be non-negligible.")
    parser.add_argument("--taxonomy", dest="taxonomy", default=False, action="store_true",
                        help="Convert phylogenetic results to taxonomic results.")
    parser.add_argument("--phyla", dest="phyla", default=False, action="store_true",
                        help="Colour phylotree results by phyla.")
    parser.add_argument("--rank", dest="rank", default=None,
                        help="Rank at which to sort results.")
    parser.add_argument("--keep-zeros", dest="keep_zeros", default=False, action="store_true",
                        help="Keep nodes of output where no reads have been assigned.")
    parser.add_argument("--ignore-names", dest="ignore_names", default=False, action="store_true")
    parser.add_argument("--group", dest="groups", action="append", nargs="+",
                        help="Space-separated list of sample files to be treated as a single group in phylotree.")
    parser.add_argument("--colour-list", dest="colour_list", nargs="+",
                        help="List of colours to use when plotting groups in phylotree.")
    parser.add_argument("--sourmash", dest="use_sourmash", default=False, action="store_true",
                        help="Use sourmash for distance estimation.")
    parser.add_argument("--rapidnj", dest="use_rapidnj", default=True, action="store_true",
                        help="Use RapidNJ for Neighbour-Joining algorithm.")
    parser.add_argument("--quicktree", dest="use_quicktree", default=False, action="store_true",
                        help="Use QuickTree for Neighbour-Joining algorithm.")
    parser.add_argument("--paired", dest="paired_end", default=False, action="store_true",
                        help="Treat reads as paired-end.")
    parser.add_argument("--alpha", dest="alpha", default=0.1,
                        help="Percentage requirement for classification subtrees (see Tutorials 1 & 2).")
    parser.add_argument("--log-scores", dest="log_scores", default=False, action="store_true",
                        help="Log transformation to opacity scores on phylotree (think uneven distributions).")
    parser.add_argument("--itol", dest="itol_mode", default=False, action="store_true",
                        help="Output plotting data in ITOL format.")
    parser.add_argument("--flat-colour", dest="flat_colour", default=False, action="store_true",
                        help="Do not use abundance to make phylotree colours opaque.")

    # Parse arguments.
    args: Namespace = parser.parse_args()

    if args.db_name is None:
        args.db_name = db_name_from_environment()
    else:
        args.db_name = make_path_absolute(args.db_name, os.getcwd())

    return ExpamOptions(**{field: getattr(args, field) for field in ExpamOptions._fields})


def db_name_from_environment():
    try:
        return os.environ["EXPAM_DB_DIR"]
    except KeyError:
        return None


def clear_logs(log_path) -> None:
    print("Clearing old log files...")

    try:
        shutil.rmtree(log_path)
    except FileNotFoundError:
        pass

    try:
        os.mkdir(log_path)
    except OSError:
        die("Can't make log path %s!" % log_path)


class CommandGroup:
    commands: Set[str] = {}

    @classmethod
    def take_args(cls, args: ExpamOptions) -> dict:
        return {}

    def run(self, command) -> None:
        if not hasattr(self, command):
            raise AttributeError("Command %s not found!" % command)
        else:
            getattr(self, command)()

    @classmethod
    def parse_ints(cls, *params):
        if len(params) == 1:
            return parse_int(params[0])
        else:
            return (parse_int(param) for param in params)

    @classmethod
    def parse_floats(cls, *params):
        if len(params) == 1:
            return parse_float(params[0])
        else:
            return (parse_float(param) for param in params)

    @staticmethod
    def get_user_confirmation(msg):
        if "y/n" not in msg:
            msg += " (y/n)"

        while True:
            ans = input(msg)
            ans = ans.lower().strip()

            if ans == "y":
                return True
            elif ans == "n":
                return False
            else:
                die("Invalid response, please provide y/n.")

    @staticmethod
    def get_time(data) -> datetime.datetime:
        date_data, time_data, am_pm = data.split(" ")[:3]

        mon, day, year = (int(v) for v in date_data.split("/"))
        hr, min, sec = (int(v) for v in time_data.split(":"))

        if "PM" in am_pm and hr < 12:
            hr += 12
        elif "AM" in am_pm and hr == 12:
            hr = 0

        return datetime.datetime(year, mon, day, hr, min, sec)

    def get_t0(self, logs_dir: str):
        def _first(itr):
            try:
                return itr[0]
            except IndexError:
                raise ValueError("Can't find main log file!")

        # Get effective t0 from main log.
        main_log_dir = _first([
            os.path.join(logs_dir, log_name)
            for log_name in ls(logs_dir, ext=".log")
            if "_main" in log_name
        ])

        with open(main_log_dir, "r") as f:
            log_data = f.readline()  # Get time of first log point.

        return main_log_dir, self.get_time(log_data)


class PlotLogs:
    def __init__(self, logs_dir: str, t0: datetime.datetime) -> None:
        self.logs_dir = logs_dir
        self.t0 = t0

    def plot(self):
        job_type_map = {  # For labels.
            "request_extension": "remap",
            "disjoint": "disjoint",
            "collapse_pile": "fill",
            "import_sequence": "kmers",
            "send_kmers": "send",
        }
        job_colours = {  # For plotting.
            "request_extension": "red",
            "disjoint": "blue",
            "collapse_pile": "green",
            "import_sequence": "orange",
            "send_kmers": "black",
        }

        process_type_jobs = {
            "_extract": [
                "import_sequence",
                "send_kmers",
            ],
            "_union": [
                "request_extension",
                "disjoint",
                "collapse_pile"
            ],
        }

        fig, ax = plt.subplots(figsize=(20, 15))
        ax.grid()

        log_urls = [
            os.path.join(self.logs_dir, log_name)
            for log_name in ls(self.logs_dir, ext=".log")
            if "_main.log" not in log_name
        ]

        # Plot data one child process at a time.
        for log_url in log_urls:
            process_type = "_extract" if "_extract.log" in log_url else "_union"

            # Data for plotting.
            data = {job_name: ([], []) for job_name in process_type_jobs[process_type]}

            with open(log_url, "r") as f:
                log_data = f.readlines()

            # Read data line by line and append data.
            for i, line in enumerate(log_data):
                for job_name in process_type_jobs[process_type]:

                    if job_name in line:
                        data[job_name][0].append(self._get_time(line) / 3600)
                        data[job_name][1].append(self._get_data(line) / 60)

            # Plot data on same graph.
            for job_name in process_type_jobs[process_type]:
                x, y = np.array(data[job_name][0]), np.array(data[job_name][1])
                ax.plot(
                    x,
                    y,
                    label=job_type_map[job_name],
                    marker="x",
                    color=job_colours[job_name],
                )

                # Don't allow duplicate labels.
                job_type_map[job_name] = ""

        ax.set(
            xlabel="Time since start [hrs]",
            ylabel="Time taken for task [mins]",
            title="Time taken to do tasks"
        )

        ax.legend(loc='best')
        fig.savefig(os.path.join(self.logs_dir, "timing.png"))

    def _get_time(self, string):
        date, current_time, am_pm = string.split(" ")[:3]
        mon, day, yr = (int(v) for v in date.split("/"))
        hr, mn, sec = (int(v) for v in current_time.split(":"))

        if "PM" in am_pm and hr < 12:
            hr += 12
        elif "AM" in am_pm and hr == 12:
            hr = 0

        time_point = datetime.datetime(yr, mon, day, hr, mn, sec)

        return (time_point - self.t0).total_seconds()

    @staticmethod
    def _get_data(string):
        return float(
            (string.split(" took ")[1])[:-3]
        )



import logging
import os
from timeit import default_timer
from expam.database import LOG_FORMAT

class Timer:
    def __init__(self):
        self.start = None
        self.elapsed = None

    def __enter__(self):
        self.start = default_timer()
        return self

    def __exit__(self, exit_type, value, traceback):
        self.elapsed = default_timer() - self.start

    def __str__(self):
        return self.verbose()

    def __float__(self):
        return self.elapsed

    def verbose(self):
        if self.elapsed is None:
            return "<not_measured>"
        return str(self.elapsed) + "s"



def timeit(func):
    def inner(*args, **kwargs):
        with Timer() as timer:
            returns = func(*args, **kwargs)

        print(f'@timeit {func.__name__} of {func.__module__}: ', timer)
        return returns

    return inner

def log_timer(func):
    # Time function and log.
    def inner(*args, **kwargs):

        loggers = logging.root.manager.loggerDict
        my_logger = None  # Null case.

        # Appropriate logger will be identified by process id.
        pid = os.getpid()

        for name in loggers:
            if str(pid) in name:
                my_logger = loggers[name]

                break

        if my_logger is None:
            raise Exception("Logger not found!")

        with Timer() as timer:
            returns = func(*args, **kwargs)

        my_logger.info(f'@timeit {func.__name__} of {func.__module__} took {str(timer)}.')
        return returns

    return inner

def new_logger(logs_dir, logger_name):
    """
    Create a fresh Logger instance that doesn't propagate
    to the root logger. Used when a new process is created.

    :param logs_dir:
    :param logger_name:
    :return:
    """
    log_file_url = os.path.join(logs_dir, logger_name + ".log")

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False  # Don't propagate to root logger.

    if not logger.hasHandlers():  # Don't duplicate logger file handlers.
        # Let each process log to a unique file.
        fh = logging.FileHandler(log_file_url)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(LOG_FORMAT)

        logger.addHandler(fh)

    return logger



import os
from typing import List, Set, Tuple
from expam.classify import ResultsPathConfig
from expam.classify.classify import ClassificationResults, name_to_id, run_classifier
from expam.classify.config import make_results_config, validate_results_configuration
from expam.classify.taxonomy import TaxonomyNCBI
from expam.cli.main import CommandGroup, ExpamOptions, clear_logs
from expam.database import FileLocationConfig
from expam.database.config import JSONConfig, make_database_config, validate_database_file_configuration
from expam.utils import die, is_hex, make_path_absolute


class ClassifyCommand(CommandGroup):
    commands: Set[str] = {
        'classify', 'to_taxonomy'
    }

    def __init__(
        self, config: FileLocationConfig,
        files: List[str], out_dir: str,
        convert_to_taxonomy: bool, cutoff: int, cpm: float, groups: List[Tuple[str]],
        use_node_names: bool, keep_zeros: bool, plot_phyla: bool,
        colour_list: List[str], paired_end: bool, alpha: float,
        log_scores: bool, itol_mode: bool
    ) -> None:
        super().__init__()

        self.config: FileLocationConfig = config
        self.results_config: ResultsPathConfig = make_results_config(out_dir)

        self.files = files
        self.out_dir = out_dir

        self.convert_to_taxonomy = convert_to_taxonomy
        self.cutoff = cutoff
        self.cpm = cpm
        self.groups = groups

        self.use_node_names = use_node_names
        self.keep_zeros = keep_zeros
        self.plot_phyla = plot_phyla
        self.colour_list = colour_list

        self.paired_end = paired_end
        self.alpha = alpha
        self.log_scores = log_scores
        self.itol_mode = itol_mode

    @classmethod
    def take_args(cls: CommandGroup, args: ExpamOptions) -> dict:
        cutoff = cls.parse_ints(args.cutoff)
        cpm, alpha = cls.parse_floats(args.cpm, args.alpha)

        if args.out_url is None:
            die("Must supply -o/--out.")

        # Format groups.
        if args.groups is not None:
            groups = [v for v in args.groups if v]

            for i in range(len(groups)):
                if is_hex(groups[i][0]):
                    groups[i] = (groups[i][0], tuple(groups[i][1:]))
                else:
                    groups[i] = (None, tuple(groups[i]))
        else:
            groups = None

        # Check colour list.
        if args.colour_list is not None:
            colour_list = [hex for hex in args.colour_list if is_hex(hex)]

            if not colour_list:
                colour_list = None
        else:
            colour_list = None

        return {
            'config': make_database_config(args.db_name),
            'files': args.directory,
            'out_dir': args.out_url,
            'convert_to_taxonomy': args.taxonomy,
            'cutoff': cutoff,
            'cpm': cpm,
            'groups': groups,
            'use_node_names': not args.ignore_names,
            'keep_zeros': args.keep_zeros,
            'plot_phyla': args.phyla,
            'colour_list': colour_list,
            'paired_end': args.paired_end,
            'alpha': alpha,
            'log_scores': args.log_scores,
            'itol_mode': args.itol_mode
        }

    def check_database_exists(self):
        validate_database_file_configuration(self.config)

        if not os.path.exists(self.config.database_file):
            die("Database has not been built! Not found at %s." % self.config.database_file)

    """
    Run command
    ===========
    
    """
    def classify(self):
        self.check_database_exists()
        clear_logs(self.config.logs)

        # Read configuration file.
        config: JSONConfig = JSONConfig(self.config.conf)
        k, n, phylogeny_path, _, _ = config.get_build_params()
        keys_shape, values_shape = tuple(config["keys_shape"]), tuple(config["values_shape"])

        # Correct for trimmed keys_shape.
        if len(keys_shape) == 1:
            keys_shape = keys_shape + (1,)

        try:
            tax_obj = TaxonomyNCBI(self.config)
            name_to_lineage, _ = tax_obj.load_taxonomy_map(self.config)
        except OSError:
            if self.convert_to_taxonomy:
                die("First run `download_taxonomy` to collect associated taxonomy data.")
            else:
                name_to_lineage = None

        # Run expam classification.
        run_classifier(
            read_paths=self.files,
            out_dir=self.out_dir,
            db_dir=self.config.base,
            k=k,
            n=n - 1,  # Account for main process.
            phylogeny_path=phylogeny_path,
            keys_shape=keys_shape,
            values_shape=values_shape,
            logging_dir=self.config.logs,
            taxonomy=self.convert_to_taxonomy,
            cutoff=self.cutoff,
            groups=self.groups,
            cpm=self.cpm,
            use_node_names=self.use_node_names,
            phyla=self.plot_phyla,
            name_taxa=name_to_lineage,
            colour_list=self.colour_list,
            paired_end=self.paired_end,
            alpha=self.alpha,
            log_scores=self.log_scores,
            itol_mode=self.itol_mode
        )
    
    """
    Convert phylogenetic results to taxonomy
    ========================================
    
    """
    def to_taxonomy(self):
        self.check_database_exists()

        if self.out_dir is None:
            die("Require output directory (-o, --out_dir)!")
        else:
            validate_results_configuration(self.results_config)

        if not os.path.exists(self.config.taxid_lineage):
            die("Run command `download_taxonomy` first to collect taxa for your genomes!")

        config = JSONConfig(self.config.conf)
        _, _, phylogeny_path, _, _ = config.get_build_params()

        index, phylogenyIndex = name_to_id(phylogeny_path)

        tax_obj = TaxonomyNCBI(self.config)
        name_to_lineage, taxon_to_rank = tax_obj.load_taxonomy_map(self.config)

        if not os.path.exists(self.results_config.tax):
            os.mkdir(self.results_config.tax)

        results = ClassificationResults(
            index=index,
            phylogeny_index=phylogenyIndex,
            results_config=self.results_config,
            groups=self.groups,
            keep_zeros=self.keep_zeros,
            cutoff=self.cutoff,
            cpm=self.cpm,
            use_node_names=self.use_node_names,
            phyla=self.plot_phyla,
            name_taxa=name_to_lineage,
            colour_list=self.colour_list,
            log_scores=self.log_scores
        )
        results.to_taxonomy(name_to_lineage, taxon_to_rank)
    


import os
import re
import shutil
import subprocess
from tempfile import TemporaryDirectory
from typing import List, Set, Tuple, Union
from expam.classify import ResultsPathConfig
from expam.classify.classify import ClassificationResults, name_to_id
from expam.classify.config import make_results_config, validate_results_configuration
from expam.classify.taxonomy import TaxonomyNCBI
from expam.cli.main import CommandGroup, ExpamOptions
from expam.database import FileLocationConfig
from expam.database.config import JSONConfig, make_database_config, validate_database_file_configuration
from expam.sequences import format_tree_string
from expam.tree import PHYLA_COLOURS
from expam.tree.tree import RandomColour
from expam.utils import die, is_hex, ls, make_path_absolute


class TreeCommand(CommandGroup):
    commands: Set[str] = {
        'phylotree', 'draw_tree', 'tree', 
        'mashtree', 'sketch', 'distance', 'nj'
    }

    def __init__(
        self, config: FileLocationConfig,
        out_dir: str, cutoff: int, cpm: float, groups: List[Tuple[str]],
        use_node_names: bool, keep_zeros: bool, plot_phyla: bool,
        colour_list: List[str], log_scores: bool, itol_mode: bool, flat_colour: bool,
        at_rank: str, use_sourmash: bool, use_quicktree: bool
    ) -> None:
        super().__init__()

        self.config: FileLocationConfig = config
        self.results_config: ResultsPathConfig = None if out_dir is None else make_results_config(out_dir)
        self.json_conf = None
        self.out_dir = out_dir

        self.cutoff = cutoff
        self.cpm = cpm
        self.group = None if groups is None else groups[0][0]
        self.groups = groups

        self.use_node_names = use_node_names
        self.keep_zeros = keep_zeros
        self.plot_phyla = plot_phyla
        self.colour_list = colour_list

        self.log_scores = log_scores
        self.itol_mode = itol_mode
        self.flat_colour = flat_colour

        self.at_rank = at_rank

        self.use_sourmash = use_sourmash
        self.use_quicktree = use_quicktree


    @classmethod
    def take_args(cls, args: ExpamOptions) -> dict:
        cutoff, n = cls.parse_ints(args.cutoff, args.n)
        cpm = cls.parse_floats(args.cpm)

        # Format groups.
        if args.groups is not None:
            groups = [v for v in args.groups if v]

            for i in range(len(groups)):
                if is_hex(groups[i][0]):
                    groups[i] = (groups[i][0], tuple(groups[i][1:]))
                else:
                    groups[i] = (None, tuple(groups[i]))
        else:
            groups = None

        # Check colour list.
        if args.colour_list is not None:
            colour_list = [hex for hex in args.colour_list if is_hex(hex)]

            if not colour_list:
                colour_list = None
        else:
            colour_list = None

        return {
            'config': make_database_config(args.db_name),
            'out_dir': args.out_url,
            'cutoff': cutoff,
            'cpm': cpm,
            'groups': groups,
            'use_node_names': not args.ignore_names,
            'keep_zeros': args.keep_zeros,
            'plot_phyla': args.phyla,
            'colour_list': colour_list,
            'log_scores': args.log_scores,
            'itol_mode': args.itol_mode,
            'flat_colour': args.flat_colour,
            'at_rank': args.rank,
            'use_sourmash': args.use_sourmash,
            'use_quicktree': args.use_quicktree
        }

    def get_conf(self):
        if self.json_conf is None:
            self.check_database_exists()
            self.json_conf = JSONConfig(self.config.conf)
        
        return self.json_conf

    def get_group_or_die(self, group_name):
        conf: JSONConfig = self.get_conf()
        k, s, sequences = conf.group_get(group_name)

        if k is None:
            die("Parameter `k` has not been set for group %s." % group_name)
        elif s is None:
            die("Parameter `s` has not been set for group %s." % group_name)

        return k, s, sequences

    def get_n_processes(self) -> int:
        conf: JSONConfig = self.get_conf()
        n_processes: Union[None, int] = conf.get_n_processes()

        if n_processes is None:
            return 1
        else:
            return int(n_processes)

    def check_database_exists(self):
        validate_database_file_configuration(self.config)

    def check_results_exist(self):
        if self.results_config is None:
            die("Results not specified!")

        if not validate_results_configuration(self.results_config):
            die("Invalid results configuration!")

    @staticmethod
    def _check_command(command):
        __has = shutil.which(command) is not None
        if not __has:
            die('Could not find local installation of %s!' % command)

    def check_mash(self):
        self._check_command('mash')

    def check_mashtree(self):
        self._check_command('mashtree')

    def check_rapidnj(self):
        self._check_command('rapidnj')

    def check_quicktree(self):
        self._check_command('quicktree')

    @staticmethod
    def shell(command, cwd=None):
        if cwd is None:
            cwd = os.getcwd()

        p = subprocess.run(
            command,
            shell=True,
            capture_output=True,
            cwd=cwd
        )
        if p.returncode != 0:
            raise Exception("\n" + p.stderr.decode("utf-8"))

        return p.stdout.decode("utf-8")
    
    """
    Make phylotree of classification output
    =======================================
    
    """
    def phylotree(self):
        self.check_results_exist()
        config: JSONConfig = self.get_conf()
        _, _, phylogeny_path, _, _ = config.get_build_params()
        index, phylogenyIndex = name_to_id(phylogeny_path)

        try:
            tax_obj: TaxonomyNCBI = TaxonomyNCBI(self.config)
            name_to_lineage, _ = tax_obj.load_taxonomy_map()
        except FileNotFoundError:
            name_to_lineage = None

        # Load phylogenetic results.
        results = ClassificationResults(
            index=index,
            phylogeny_index=phylogenyIndex,
            results_config=self.results_config,
            groups=self.groups,
            keep_zeros=self.keep_zeros,
            cutoff=self.cutoff,
            cpm=self.cpm,
            use_node_names=self.use_node_names,
            phyla=self.plot_phyla,
            name_taxa=name_to_lineage, 
            colour_list=self.colour_list,
            log_scores=self.log_scores
        )
        results.draw_results(itol_mode=self.itol_mode, flat_colour=self.flat_colour)

    """
    Draw phylogenetic tree
    ======================
    
    """
    def draw_tree(self):
        try:
            from ete3 import AttrFace, faces, Tree, TreeStyle, NodeStyle
        except ImportError:
            die("First install ete3 tools...`pip install ete3`.")
        
        config: JSONConfig = self.get_conf()
        _, _, phylogeny_path, _, _ = config.get_build_params()

        if phylogeny_path is None:
            die("Phylogeny not set!")

        try:
            with open(phylogeny_path, "r") as f:
                newick_string = f.read().strip()
        except OSError:
            die("Can't find Newick file %s." % phylogeny_path)

        tree = Tree(newick_string, format=1)
        rank_colours = {}

        tax_obj: TaxonomyNCBI = TaxonomyNCBI(self.config)
        name_to_lineage, name_to_rank = tax_obj.load_taxonomy_map()

        random_colour_maker: RandomColour = RandomColour()

        def layout(node):
            nonlocal name_to_lineage, name_to_rank, random_colour_maker

            if node.is_leaf():
                if self.use_node_names:
                    faces.add_face_to_node(AttrFace("name", fsize=20), node, column=0, position="aligned")

                if self.at_rank is not None:
                    lineage = name_to_lineage[node.name]

                    for name in lineage:
                        rank = name_to_rank[name][1]

                        if rank == self.at_rank:
                            if name not in rank_colours:
                                rank_colours[name] = next(random_colour_maker)

                            node_style = NodeStyle()
                            node_style['bgcolor'] = rank_colours[name]
                            node.set_style(node_style)

                elif self.plot_phyla:
                    for phyla, phyla_colour in PHYLA_COLOURS:
                        lineage = name_to_lineage[node.name]

                        if phyla in lineage:
                            node_style = NodeStyle()
                            node_style['bgcolor'] = phyla_colour
                            node.set_style(node_style)

        # Print tree to file.
        ts = TreeStyle()
        ts.mode = "c"
        ts.show_leaf_name = False
        ts.layout_fn = layout
        ts.force_topology = True
        ts.allow_face_overlap = False
        ts.draw_guiding_lines = True
        ts.root_opening_factor = 1

        phy_path = "phylotree.pdf" if not self.out_dir else os.path.join(self.out_dir, "phylotree.pdf")
        tree.render(
            phy_path,
            tree_style=ts,
            w=4000,
            units="px"
        )
        print("Phylogenetic tree written to %s!" % phy_path)

    """
    CLI for making phylogenetic tree
    ================================
    
    """
    def tree(self):
        conf: JSONConfig = self.get_conf()

        for task, task_check in [
            (self.do_sketches, self.check_sketches),
            (self.do_distances, self.check_distances),
            (self.do_trees, self.check_trees)
        ]:
            if not task_check():
                task()

        tree_dir = self.finalise_tree()
        conf.set(phylogeny_path=tree_dir)
        conf.save()

        print("Phylogeny set to %s." % tree_dir)


    def finalise_tree(self):
        print("Finalising tree...")
        conf: JSONConfig = self.get_conf()

        tree_dir = os.path.join(self.config.phylogeny, 'tree')
        template_name = "template.nwk"
        tree_name = "%s.nwk" % os.path.basename(self.config.phylogeny.rstrip(os.sep))
        template_path = os.path.join(tree_dir, template_name)
        tree_path = os.path.join(tree_dir, tree_name)
        tree_files = [
            name
            for name in ls(tree_dir, ext='.nwk')
            if not name.endswith(tree_name)
        ]

        def get_tree_data(file_url):
            # read tree data
            with open(file_url, 'r') as f:
                tree_data = f.read().strip().rstrip(";")
            # format tree node names
            tree_data = format_tree_string(tree_data)
            return tree_data

        if len(tree_files) == 1:
            nwk_data = get_tree_data(tree_files[0]).rstrip(";") + ";"
            with open(tree_path, 'w') as f:
                f.write(nwk_data)
        else:
            try:
                template = get_tree_data(template_path) + ";"
            except FileNotFoundError:
                die("No template found! Please write a template to %s!\n See expam.readthedocs.io/en/latest/tutorials/treebuilding.html#part-two-building-a-tree-in-parts for details." % template_path)

            template_groups = re.findall(r"{{(\S+?)}}", template)
            tree_data = {
                group_name: get_tree_data(tree_file)
                for group_name in conf.groups()
                for tree_file in tree_files
                if group_name in tree_file
            }

            for template_group in template_groups:
                if template_group not in tree_data:
                    die("Template group %s has no tree!" % template_group)
                else:
                    template = template.replace("{{%s}}" % template_group, tree_data[template_group])

            with open(tree_path, 'w') as f:
                f.write(template)

        return tree_path

    """
    Make mashtree
    =============
    
    """
    def mashtree(self):
        self.do_mashtree()
        self.tree()

    def do_mashtree(self):
        print("Creating mashtree...")
        self.check_mashtree()
        conf = self.get_conf()
        tree_dir = os.path.join(self.config.phylogeny, 'tree')
        temp_dir = os.path.join(self.config.phylogeny, 'tmp')
        n = self.get_n_processes()

        if not os.path.exists(tree_dir):
            os.mkdir(tree_dir)

        for group_name in conf.get_groups(self.group):
            k, s, sequences = self.get_group_or_die(group_name)
            tree_path = os.path.join(tree_dir, "%s.nwk" % group_name)
            self.make_mashtree(k, s, n, sequences, tree_path, temp_dir)

    def make_mashtree(self, k, s, n, sequences, tree_dir, temp_dir):
        _names_file = os.path.join(temp_dir, 'sequence_names.txt')

        def delete_temp():
            nonlocal temp_dir
            shutil.rmtree(temp_dir)
            print("Deleted temporary directory %s." % temp_dir)

        # Make temporary directory.
        try:
            print("Making temporary directory %s..." % temp_dir)
            os.mkdir(temp_dir)
        except OSError:
            try:
                delete_temp()
            except FileNotFoundError:
                pass
            die("Could not create temporary directory!")

        # Write file of files.
        try:
            with open(_names_file, "w") as f:
                f.write("\n".join(sequences))

        except OSError:
            die("Could not write genome paths to text file!")

        # Make mashtree and insert into configuration file.
        print("Making mashtree...")
        cmd = "mashtree --numcpus %d --kmerlength %d --sketch-size %d --file-of-files %s" \
            % (n, k, s, _names_file)
        try:
            return_val = self.shell(cmd, cwd=temp_dir)
            with open(tree_dir, "w") as f:
                f.write(return_val)
            print("Phylogeny created in %s." % tree_dir)
        finally:
            try:
                delete_temp()
            except FileNotFoundError:
                pass
            except OSError:
                die("Could not delete temporary directory %s!" % temp_dir)

    """
    Sketch sequences
    ================
    
    """
    def sketch(self):
        self.do_sketches()

    def check_sketches(self):
        conf: JSONConfig = JSONConfig(self.config.conf)
        sketch_dir = os.path.join(self.config.phylogeny, 'sketch')
        sketch_name_fmt = "%s.k%d.s%d.%s"

        if not os.path.exists(sketch_dir):
            return False

        for group_name in conf.get_groups(self.group):
            k, s, _ = self.get_group_or_die(group_name)
            file_name = sketch_name_fmt % (group_name, k, s, "%s")

            dest = os.path.join(sketch_dir, file_name % ("sour" if self.use_sourmash else "msh"))
            if not os.path.exists(dest):
                return False

        return True

    def do_sketches(self):
        conf: JSONConfig = JSONConfig(self.config.conf)
        n: int = self.get_n_processes()

        sketch_dir = os.path.join(self.config.phylogeny, 'sketch')
        if not os.path.exists(sketch_dir):
            os.mkdir(sketch_dir)

        sketch_name_fmt = "%s.k%d.s%d.%s"

        for group_name in conf.get_groups(self.group):
            k, s, sequences = self.get_group_or_die(group_name)
            file_name = sketch_name_fmt % (group_name, k, s, "%s")

            if self.use_sourmash:
                sig_path = os.path.join(sketch_dir, file_name % "sour")
                if not os.path.exists(sig_path):        
                    os.mkdir(sig_path)

                self.sour_sketch(k=k, s=s, sequences=sequences, sig_dir=sig_path)

            else:
                self.check_mash()

                file_path = os.path.join(sketch_dir, file_name % "msh")
                self.mash_sketch(k=k, s=s, p=n, sequences=sequences, out_dir=file_path)

    def sour_sketch(self, k: int, s: int, sequences: List[str], sig_dir: str):
        from expam.tree.sourmash import make_signatures
        make_signatures(self.get_n_processes(), sequences, sig_dir, k, s)

    def mash_sketch(self, k, s, p, sequences, out_dir):
        # prepare temporary directory to link files
        with TemporaryDirectory() as tmp_dir:
            for seq in sequences:
                os.symlink(seq, os.path.join(tmp_dir, os.path.basename(seq)))
            # run mash on temp directory
            cmd_fmt = "mash sketch -k %d -p %d -s %d -o %s %s/*"
            cmd = cmd_fmt % (k, p, s, out_dir, tmp_dir)
            self.shell(cmd)

    """
    Compute pairwise distances between sketched sequences
    =====================================================
    
    """
    def distance(self):
        self.do_distances()

    def check_distances(self):
        conf: JSONConfig = self.get_conf()
        dist_dir = os.path.join(self.config.phylogeny, 'distance')

        if not os.path.exists(dist_dir):
            return False

        dist_name_fmt = "%s.k%d.s%d.tab"

        for group_name in conf.get_groups(self.group):
            k, s, _ = self.get_group_or_die(group_name)
            file_name = dist_name_fmt % (group_name, k, s)

            dest = os.path.join(dist_dir, file_name)
            if not os.path.exists(dest):
                return False

        return True

    def do_distances(self):
        print("Calculating pairwise distances...")
        conf: JSONConfig = self.get_conf()

        sketch_base = os.path.join(self.config.phylogeny, 'sketch')
        dist_base = os.path.join(self.config.phylogeny, 'distance')

        if not os.path.exists(dist_base):
            os.mkdir(dist_base)

        dist_name_fmt = "%s.k%d.s%d.%s"

        for group_name in conf.get_groups(self.group):
            k, s, _ = self.get_group_or_die(group_name)
            sketch_name = dist_name_fmt % (group_name, k, s, '%s')

            matrix_name = dist_name_fmt % (group_name, k, s, 'tab')
            matrix_file = os.path.join(dist_base, matrix_name)

            if self.use_sourmash:
                sketch_dir = os.path.join(sketch_base, sketch_name % 'sour')
                self.sour_dist(sig_dir=sketch_dir, matrix_dir=matrix_file)
            else:
                self.check_mash()

                sketch_file = os.path.join(sketch_base, sketch_name % 'msh')
                self.mash_dist(sketch_file, matrix_file)

    def mash_dist(self, sketch_dir, matrix_dir):
        cmd_fmt = "mash dist -p %d -t %s %s"
        cmd = cmd_fmt % (
            self.get_n_processes(),
            sketch_dir,
            sketch_dir
        )

        _unfmt_dst = self.shell(cmd)
        lines = [line.strip() for line in _unfmt_dst.split('\n')]

        names = lines[0].split('\t')
        lines[0] = '%d' % len(names[1:])

        # Format sequences names from within directories.
        for i in range(1, len(lines)):
            line = lines[i].split('\t')
            line[0] = os.path.basename(line[0])
            lines[i] = '\t'.join(line)

        with open(matrix_dir, 'w') as f:
            f.write('\n'.join(lines))

    def sour_dist(self, sig_dir, matrix_dir):
        from expam.tree.sourmash import make_distances
        make_distances(sig_dir, self.get_n_processes(), matrix_dir)

    """
    Execute neighbour-joining on distance matrix
    ============================================
    
    """
    def nj(self):
        self.do_trees()

    def check_trees(self):
        conf: JSONConfig = self.get_conf()

        tree_dir = os.path.join(self.config.phylogeny, 'tree')
        if not os.path.exists(tree_dir):
            return False

        for group_name in conf.get_groups(self.group):
            tree_name = '%s.nwk' % group_name
            tree_path = os.path.join(tree_dir, tree_name)

            if not os.path.exists(tree_path):
                return False

        return True

    def do_trees(self):
        print("Running neighbour-joining...")
        conf: JSONConfig = self.get_conf()

        dist_dir = os.path.join(self.config.phylogeny, 'distance')
        tree_dir = os.path.join(self.config.phylogeny, 'tree')
        dist_fmt = "%s.k%d.s%d.tab"

        if not os.path.exists(tree_dir):
            os.mkdir(tree_dir)

        for group_name in conf.get_groups(self.group):
            k, s, _ = self.get_group_or_die(group_name)

            matrix_name = dist_fmt % (group_name, k, s)
            matrix_path = os.path.join(dist_dir, matrix_name)

            tree_name = "%s.nwk" % group_name
            tree_path = os.path.join(tree_dir, tree_name)

            if self.use_quicktree:
                self.quicktree(matrix_dir=matrix_path, tree_dir=tree_path)
            else:
                self.rapidnj(matrix_dir=matrix_path, tree_dir=tree_path)

    def rapidnj(self, matrix_dir, tree_dir):
        self.check_rapidnj()

        cmd_fmt = "rapidnj %s -i pd -o t -c %d"
        cmd = cmd_fmt % (matrix_dir, self.get_n_processes())

        unformatted_tree = self.shell(cmd)
        tree = self._format_tree_string(unformatted_tree)

        with open(tree_dir, 'w') as f:
            f.write(tree)

        print("RapidNJ wrote tree to %s." % tree_dir)

    def quicktree(self, matrix_dir, tree_dir):
        self.check_quicktree()

        cmd_fmt = "quicktree -in m -out t %s"
        cmd = cmd_fmt % matrix_dir

        unformatted_tree = self.shell(cmd)
        tree = self._format_tree_string(unformatted_tree)

        with open(tree_dir, 'w') as f:
            f.write(tree)

        print("QuickTree wrote tree to %s." % tree_dir)

    @staticmethod
    def _format_tree_string(nwk):
        seq_types = [".fna", ".faa"]
        comp_types = [".tar.gz", ".gz"]

        nwk = nwk.replace("'", "")

        for seq_type in seq_types:
            for comp_type in comp_types:
                ext = seq_type + comp_type + ":"
                nwk = nwk.replace(ext, ":")

        return nwk



from multiprocessing import Process
import queue
import time
import traceback

import psutil

from expam.database import TIMEOUT
from expam.logger import new_logger
from expam.process import COMMAND_CODES


class JobWorker(Process):
    def __init__(
        self, logging_dir, status_index, statuses,
        child_statuses, timeout=TIMEOUT, job_name="Worker",
        phases=None
    ):
        super(JobWorker, self).__init__()

        self._process = None

        self._status_id = status_index
        self._fellow_statuses = statuses
        self._child_statuses = child_statuses

        self.phases = [] if phases is None else list(phases)
        self.work_queue = None  # Where work comes from.
        self.child_queues = None  # Where to distribute work.

        #
        # Defined at ControlCenter.
        self._parents = []
        self._children = []
        self._command_permissions = {
            "command": "None",
            "counts": 0
        }

        #
        # User defined methods.
        self._getters = {}
        self._processors = {}
        self._transitions = {}
        self._shutdown = self.default_shutdown

        self._timeout = timeout
        self.job_name = job_name

        # Logging to disk.
        self.logging_dir = logging_dir
        self.logger = None

    def set_method(self, method_name, phase, method):
        methods = {
            "getter": self._getters,
            "processor": self._processors,
            "transition": self._transitions,
        }

        if method_name not in methods:
            raise ValueError("Invalid method name %s!" % method_name)

        if method is not None:
            methods[method_name][phase] = method

    def init_methods_from_dict(self, config):

        if not isinstance(config, dict):
            raise TypeError("Config must be dictionary not %s!" % str(type(config)))

        for key, value in config.items():
            if value is None:
                continue

            if key == "shutdown":

                if not callable(value):
                    raise TypeError("Shutdown object must be callable!")

                self.set_shutdown(value)

            elif key in self.phases:

                if not isinstance(value, dict):
                    raise TypeError("Method declaration must be dictionary not %s!"
                                    % str(type(value)))

                for method_name, method in value.items():
                    self.set_method(
                        method_name=method_name,
                        phase=key,
                        method=method
                    )

            else:
                raise ValueError("Unknown key in class config: %s!" % key)

    def set_shutdown(self, method):
        self._shutdown = method

    def children_have_obeyed(self, *codes):
        if self._child_statuses is None:
            return True

        return all([
            child_status in codes
            for child_status in self._child_statuses
        ])

    @property
    def is_primary_child(self):
        return True if self._status_id == 0 else False

    @property
    def children_have_errors(self):
        if self._child_statuses is None:
            return False

        return any([
            child_status == COMMAND_CODES["ERROR"]
            for child_status in self._child_statuses
        ])

    @property
    def error_propagated(self):
        if self._child_statuses is None:
            return True

        return all([
            child_status == COMMAND_CODES["ERROR"]
            for child_status in self._child_statuses
        ])

    @property
    def children_dead(self):
        return self.children_have_obeyed(COMMAND_CODES["DEAD"])

    def peers_are_okay(self, cid):
        return all([
            peer_status != COMMAND_CODES["ERROR"]
            for i, peer_status in enumerate(self._fellow_statuses)
            if i != cid
        ])

    def wake_children(self):
        for cid in self.yield_children_id():
            self._child_statuses[cid] = COMMAND_CODES["ACTIVE"]

    @property
    def child_id(self):
        return self._status_id

    @property
    def n_parents(self):
        return len(self._parents)

    @property
    def n_children(self):
        return len(self._children)

    @property
    def n_peers(self):
        return len(self._fellow_statuses)

    @property
    def phase(self):
        return self.phases[0]

    @property
    def status(self):
        return self._fellow_statuses[self._status_id]

    @property
    def name(self):
        return self.job_name

    @status.setter
    def status(self, value):
        assert value in COMMAND_CODES.values()

        self._fellow_statuses[self._status_id] = value

    @property
    def current_command(self):
        return self._command_permissions["command"]

    @current_command.setter
    def current_command(self, value):
        self._command_permissions["command"] = value

    @property
    def command_permissions(self):
        return self._command_permissions["counts"]

    @command_permissions.setter
    def command_permissions(self, value):
        self._command_permissions["counts"] = value

    @property
    def is_supposed_to_die(self):
        return self.status == COMMAND_CODES["ERROR"]

    @property
    def is_supposed_to_sleep(self):
        return self.status == COMMAND_CODES["SLEEP"]

    def reset_command_permissions(self):
        self.current_command = "None"
        self.command_permissions = 0

    def tell_children(self, *args):
        if len(args) == 1:
            args = args * self.n_children

        assert len(args) == self.n_children

        for i in range(self.n_children):
            self.tell_child(i, args[i])

    def tell_child(self, cid, data):

        if not self.child_is_dead(cid):
            self.child_queues[cid].put(data)

    def message_child(self, cid, err_data):
        self._children[cid].send(err_data)

    def yield_children_id(self):
        for i in range(len(self._children)):
            yield i

    def timeout(self):
        time.sleep(self._timeout)

    def is_active(self):
        if self.status == COMMAND_CODES["ACTIVE"]:
            return True
        return False

    def _increment_command(self, command_string):

        if self.current_command == "None":
            self.current_command = command_string

        elif self.current_command != command_string:
            raise ValueError("New command %s interrupted old command flow (%s)!"
                             % (command_string, self.current_command))

        # Check completion condition.
        self.command_permissions += 1

        if self.command_permissions == self.n_parents:

            command_code = COMMAND_CODES[command_string]
            death_code = COMMAND_CODES["DEAD"]

            # Broadcast sleep command to all children.
            self.log("%d moving command to children..." % self.pid)
            self.tell_children((command_code, command_string))
            self.log("%d children told!" % self.pid)

            # Children must have either slept or have died due to no more phases
            while not self.children_have_obeyed(command_code, death_code):
                if self.is_primary_child:
                    has_errors = self.check_children()  # Monitor for errors.

                    if has_errors:
                        return

                # Listen for messages.
                messages = self.check_for_messages()
                if messages:
                    self.begin_error_sequence(messages)

                    return

                self.timeout()

            self.log("%d my children have obeyed!" % self.pid)

            # Start transition to sleep state.
            self.reset_command_permissions()
            self.status = COMMAND_CODES["TRANSITION"]

    def activate(self):
        if not self.is_active():
            self.status = COMMAND_CODES["ACTIVE"]

    def process(self, *args):
        memory_info = self._process.memory_info()
        self.log("%d currently using (RSS: %fGb, VSZ: %fGb) of RAM. Processing data..."
                 % (self.pid,
                    memory_info.rss / 1024 ** 3,
                    memory_info.vms / 1024 ** 3))

        return self._processors[self.phase](*args)

    def sleep(self, *args):
        self.log("%d received sleep command." % self.pid)

        self._increment_command("SLEEP")

        if self.status == COMMAND_CODES["TRANSITION"]:
            self.transition(*args)

            # Sleep when phase transition is complete.
            self.status = COMMAND_CODES["SLEEP"]

            # Wait to be awoken by parent.
            while not self.is_active():
                self.timeout()

            # Wake up children.
            self.wake_children()

    def die(self, *args):
        self.log("%d received die command." % self.pid)

        self._increment_command("DEAD")

        if self.status == COMMAND_CODES["TRANSITION"]:
            self.shutdown(*args)

            # Die when phase transition is complete.
            self.status = COMMAND_CODES["DEAD"]

    def set_error_status(self):
        self.status = COMMAND_CODES["ERROR"]

    def reset_status(self):
        self.status = COMMAND_CODES["ACTIVE"]

    def set_dead_status(self):
        self.status = COMMAND_CODES["DEAD"]

    def die_error(self, *args):
        self.set_error_status()

        # Clear my work queue to shut down safely.
        self.clear_work_queue()

        self.shutdown()

    def clear_work_queue(self):
        while True:
            try:
                _ = self.work_queue.get(block=True, timeout=self._timeout)

            except queue.Empty:
                return

    def get_work(self, *args):

        # First check for any errors raised.
        msg = self.check_for_messages()

        if msg is not False:  # This is an error to be propagated.
            return msg

        else:
            if self.phases and self.phase in self._getters:
                return self._getters[self.phase](*args)

            assert len(args) == 2
            wait, pid = args[0], args[1]

            # Check for more jobs.
            try:
                work = self.work_queue.get(block=True, timeout=self._timeout)

                return work

            except queue.Empty:
                if not wait:
                    return None

    def loop_condition(self, cond):
        if self.received_message():
            return False

        return cond

    def received_message(self):
        if self._parents[0].poll(self._timeout):
            return True

        return False

    def check_for_messages(self):
        if self._parents[0].poll(self._timeout):
            return self._parents[0].recv()

        else:
            return False

    def transition(self, *args):
        if self.phases and self.phase in self._transitions:
            self._transitions[self.phase](*args)

        self.phases.pop(0)
        self.log("%d now at phase %s!" % (self.pid, ("end" if not self.phases else self.phase)))

    def shutdown(self, *args):
        return self._shutdown(*args)

    def default_shutdown(self, *args):
        self.log("%d %s closing..." % (self.pid, self.job_name))

    def raise_error(self, message):
        # If we already have an error message, that takes precedence.
        # Only send if there are no current messages (ie. peers haven't raised).

        self.set_error_status()

        if self._parents[0].poll(self._timeout):
            self.log("%d error was overridden!" % self.pid)

            # Another error was raised.
            self.begin_error_sequence()

        else:
            self._parents[0].send(message)
            self.log("%d error raised..." % self.pid)

            # Await confirmation that my error was received.
            self.log("%d awaiting response..." % self.pid)
            receipt = self._parents[0].recv()
            self.log("%d response received!" % self.pid)

            if receipt is True:
                self.log("%d sent error message!" % self.pid)

                if self.is_primary_child:
                    self.propagate_error(-1, message)

                self.die_error(message[1])

            else:
                self.log("%d error was overridden!" % self.pid)

                # Another error was raised first.
                self.begin_error_sequence(receipt)

    def propagate_error(self, eid, err_data):
        self.log("%d propagating error..." % self.pid)

        # Tell children about error and wait for them to shut down.
        for cid in self.yield_children_id():
            if cid != eid:
                self.message_child(cid, err_data)

        while not self.children_dead:
            self.timeout()

        self.log("%d error propagated!" % self.pid)

    def check_children(self):
        assert self.is_primary_child

        if self.children_have_errors:
            self.log("%d found errors!" % self.pid)
            self.set_error_status()

            # Get error and data.
            for cid in self.yield_children_id():

                if self._children[cid].poll(self._timeout):
                    err_data = self._children[cid].recv()
                    self._children[cid].send(True)  # Send confirmation.

                    break

            else:
                raise Exception("No errors found even though children are dead!")

            self.propagate_error(cid, err_data)

            # Tell parent.
            self.raise_error(err_data)

            return True

        return False

    def child_is_dead(self, cid):
        if self._child_statuses[cid] == COMMAND_CODES["DEAD"]:
            return True

        return False

    def begin_error_sequence(self, err_data=None):
        if err_data is None:
            err_data = self._parents[0].recv()

        if self.is_primary_child:
            self.propagate_error(-1, err_data)

        self.die_error(err_data)

    def log(self, msg):
        self.logger.info(msg)

    def run(self):
        # Activate logging.
        self.logger = new_logger(self.logging_dir, str(self.pid) + "_" + self.job_name)
        self.log("%d %s is now awake!" % (self.pid, self.job_name))

        # Activate memory tracking.
        self._process = psutil.Process(self.pid)

        # Phases loop.
        while not self.is_supposed_to_die and self.phases:

            # Check for errors.
            if self.is_primary_child:
                has_errors = self.check_children()

                if has_errors:
                    break

            # Processing loop.
            for pid in range(self.n_parents):

                try:
                    next_job = self.get_work(False, pid)  # True ==> wait for work.

                    if next_job is None:
                        break

                    command_code, data = next_job

                    if command_code == COMMAND_CODES["ACTIVE"]:
                        self.process(data)

                    elif command_code == COMMAND_CODES["SLEEP"]:
                        self.sleep(data)

                        if self.is_supposed_to_sleep:
                            break  # Check next phase exists.

                    elif command_code == COMMAND_CODES["DEAD"]:
                        self.die(data)

                        if self.is_supposed_to_die:
                            break  # Check next phase exists.

                    elif command_code == COMMAND_CODES["ERROR"]:
                        # Error message is currently propagating.
                        self.begin_error_sequence((command_code, data))

                        break

                    else:
                        raise ValueError("Invalid command code %d!" % command_code)

                except Exception as e:
                    # Log error with context.
                    self.log("*** Exception raised! ***\n" + traceback.format_exc())

                    self.log("Exception encountered in %d! Attempting raise..." % self.pid)

                    err_data = (COMMAND_CODES["ERROR"], e)
                    self.raise_error(err_data)

                    break

        if not self.is_supposed_to_die:
            self.shutdown()

        self.set_dead_status()




import json
import os
from posixpath import isabs

import numpy as np
from expam.database import ACCESSION_ID_RELATIVE_PATH, CONF_RELATIVE_PATH, DATABASE_FILE_RELATIVE_PATH, DATABASE_RELATIVE_PATH, LCA_MATRIX_RELATIVE_PATH, LOG_RELATIVE_PATH, PHYLOGENY_RELATIVE_PATH, TAXID_LINEAGE_MAP_RELATIVE_PATH, TAXON_RANK_MAP_RELATIVE_PATH, FileLocationConfig
from expam.utils import die


class JSONConfig:
    _parse_data_type = {
        "uint8": np.uint8,
        "uint16": np.uint16,
        "uint32": np.uint32,
        "uint64": np.uint64,
        "int8": np.int8,
        "int16": np.int16,
        "int32": np.int32,
        "int64": np.int64,
    }

    def __init__(self, url=None, name: str = ""):
        self.url = url

        # Default values.
        self.params = {
            "name": name,
            "groups": {
                'default': {
                    'sequence_files': [],
                    'phylogeny_path': [],
                    'k': None,
                    's': None,
                },
            },
            "phylogeny_path": None,
            "k": None,
            "n": None,
            "s": None,
            "pile": None,
            "keys_shape": None,
            "keys_data_type": None,
            "values_shape": None,
            "values_data_type": None,
        }

        if url is not None:
            with open(url, 'r') as f:
                _conf = json.load(f)

                if not isinstance(_conf, dict):
                    die('Invalid configuration file!')

                for key in _conf:
                    if key == 'groups':
                        self.params['groups'].update(_conf['groups'])

                    else:
                        self.params[key] = _conf[key]

    def __getitem__(self, key):
        return self.params[key]

    def __str__(self):
        name_line = "<<< expam configuration file: %s >>>\n" % self.params['name']
        data_line = "phylogeny\t-->\t%s\nk\t\t-->\t%s\nn\t\t-->\t%s" % (self.params['phylogeny_path'],
                                                                    self.params['k'], self.params['n'])
        par_line = "sketch\t\t-->\t%s\npile\t\t-->\t%s" % (self.params['s'], self.params['pile'])

        group_fmt = "\n----------------\ngroup name: %s\n\tk\t\t-->\t%s\n\tsketch\t\t-->\t%s\n\tsequences\t-->\t%d\n"
        group_line = "\n".join([
            group_fmt % (name, group['k'], group['s'], len(group['sequence_files']))
            for name, group in self.params['groups'].items()
        ])

        return "\n".join([name_line, data_line, par_line, group_line])

    def get_paths(self):
        return [
            genome_path
            for group in self.params["groups"].values()
            for genome_path in group['sequence_files']
        ]

    def new_group(self, name, files=None, phylogeny_path=None, k=None, s=None):
        if name in self.params['groups']:
            raise ValueError("Group %s already exists!" % name)

        self.params['groups'][name] = {
            'sequence_files': [] if files is None else files,
            'phylogeny_path': phylogeny_path,
            'k': k,
            's': s,
        }

    def set(self, **kwargs):
        for arg in kwargs:
            if arg in self.params and arg != 'groups' and arg != 'phylogeny_path':
                if kwargs[arg] is not None:
                    self.params[arg] = kwargs[arg]
            elif arg == "phylogeny_path":
                phylogeny_path = kwargs[arg]

                if phylogeny_path is not None:
                    if not os.path.isabs(phylogeny_path):
                        abs_path = os.path.abspath(phylogeny_path)

                        if not os.path.exists(abs_path) and self.url is not None:
                            db_path = self.url.rstrip(CONF_RELATIVE_PATH)
                            phy_db_rel_path = os.path.join(db_path, phylogeny_path)

                            if not os.path.exists(phy_db_rel_path):
                                die("Can't find path %s." % kwargs[arg])
                        elif not os.path.exists(abs_path) and self.url is None:
                            die("Can't find path %s." % kwargs[arg])
                        else:
                            phylogeny_path = abs_path
                
                    self.params[arg] = phylogeny_path
            else:
                raise ValueError("Unknown configuration category %s!" % arg)

    def group_set(self, group_name, **kwargs):
        for arg in kwargs:
            if arg in self.params['groups'][group_name] and arg != 'sequence_files' and kwargs[arg] is not None:
                self.params['groups'][group_name][arg] = kwargs[arg]

    def group_get(self, group_name):
        if group_name not in self.params['groups']:
            raise IndexError("Invalid group %s!" % group_name)

        group = self.params['groups'][group_name]
        attrs = {attr: group[attr] for attr in ('k', 's', 'sequence_files')}

        for req_param in ('k', 's'):
            if group[req_param] is None:
                if self.params[req_param]:
                    attrs[req_param] = int(self.params[req_param])

            else:
                attrs[req_param] = int(attrs[req_param])

        return attrs['k'], attrs['s'], attrs['sequence_files']

    def add_sequence(self, file_name, group=None):
        if group is None:
            group = 'default'

        if group not in self.params['groups']:
            self.new_group(group)

        if not isinstance(file_name, list):
            file_name = {file_name}
        else:
            file_name = set(file_name)

        # Check these sequences aren't already added.
        for current_group in self.params['groups']:
            if current_group != group:
                current_seqs = set(self.params['groups'][current_group]['sequence_files'])

                if file_name & current_seqs:
                    die("Some of these sequences are already in the database!")

        self.params['groups'][group]['sequence_files'].extend(file_name)

    def remove_sequence(self, file_name, group=None):
        if group is None:
            group = 'default'

        if isinstance(file_name, list):
            file_name = set(file_name)
        else:
            file_name = {file_name}

        sequences = set(self.params['groups'][group]['sequence_files'])
        sequences -= file_name

        self.params['groups'][group]['sequence_files'] = list(sequences)

    def remove_group(self, group=None):
        if group is None:
            self.params['groups']['default'] = {
                    'sequence_files': [],
                    'phylogeny_path': [],
                    'k': None,
                    's': None,
                }
        elif group in self.params['groups']:
            self.params['groups'].pop(group)
        else:
            raise Exception("No such group %s!" % group)

    def _dump(self):
        return json.dumps(self.params)

    def groups(self):
        return list(self.params['groups'].keys())

    def save(self, url=None):
        if url is None:
            if self.url is None:
                raise Exception("Save url required!")

            url = self.url

        with open(url, 'w') as f:
            json.dump(self.params, f)

    def get_build_params(self):
        """Returns k, n, phylogeny_path, genome_paths, pile_size"""
        genome_paths, phylogeny_path, k, n, pile_size = self.get_paths(), self["phylogeny_path"], int(self["k"]), int(self["n"]), self["pile"]

        if not os.path.exists(phylogeny_path):
            assert self.url is not None
            db_path = self.url.rstrip(CONF_RELATIVE_PATH)
            full_path = os.path.join(db_path, phylogeny_path)

            if not os.path.join(full_path):
                die("Can't find phylogeny! Original path is %s." % phylogeny_path)
            else:
                phylogeny_path = full_path

        return k, n, phylogeny_path, genome_paths, pile_size

    def get_groups(self, input_group: str = None):
        groups = self.groups() if input_group is None else [input_group]

        # Check group has more than zero sequences.
        for group in groups:
            _, _, sequences = self.group_get(group)

            if len(sequences) > 0:
                yield group

    def get_n_processes(self):
        return self['n']


class ExpamDatabaseDoesNotExistError(Exception):
    pass

class ExpamDatabaseExistsError(Exception):
    pass


def make_database_config(db_path: str) -> FileLocationConfig:
    database_file_locations = {
        'base': db_path,
        'database': os.path.join(db_path, DATABASE_RELATIVE_PATH),
        'phylogeny': os.path.join(db_path, PHYLOGENY_RELATIVE_PATH),
        'logs': os.path.join(db_path, LOG_RELATIVE_PATH),
        'conf': os.path.join(db_path, CONF_RELATIVE_PATH),
        'accession_id': os.path.join(db_path, ACCESSION_ID_RELATIVE_PATH),
        'taxid_lineage': os.path.join(db_path, TAXID_LINEAGE_MAP_RELATIVE_PATH),
        'taxon_rank': os.path.join(db_path, TAXON_RANK_MAP_RELATIVE_PATH),
        'lca_matrix': os.path.join(db_path, LCA_MATRIX_RELATIVE_PATH),
        'database_file': os.path.join(db_path, DATABASE_FILE_RELATIVE_PATH),
    }
    return FileLocationConfig(**database_file_locations)


def load_database_config(db_path: str) -> FileLocationConfig:
    proposed_config: FileLocationConfig = make_database_config(db_path)

    validate_database_file_configuration(proposed_config)
    return proposed_config


def create_database(config: FileLocationConfig) -> None:
    for field_to_check in ('base', 'database', 'phylogeny', 'logs'):
        path: str = getattr(config, field_to_check)

        if not os.path.exists(path):
            os.mkdir(path)
        else:
            raise ExpamDatabaseExistsError("Database %s already exists!" % config.database)

    # Create new configuration file.
    conf: JSONConfig = JSONConfig()
    conf.save(url=config.conf)


def validate_database_file_configuration(proposed_config: FileLocationConfig) -> bool:
    for field_to_check in ('base', 'database', 'phylogeny', 'conf'):
        path_to_check = getattr(proposed_config, field_to_check)

        if not os.path.exists(path_to_check):
            print(path_to_check)
            raise ExpamDatabaseDoesNotExistError("Database does not exist at %s" % path_to_check)


def validate_taxonomy_files(config: FileLocationConfig) -> bool:
    for field_to_check in ('accession_id', 'taxid_lineage', 'taxon_rank'):
        if not os.path.exists(getattr(config, field_to_check)):
            return False
    else:
        return True



import os
from typing import Set
import matplotlib.pyplot as plt
from expam.classify import ResultsPathConfig
from expam.classify.config import make_results_config, validate_classification_results, validate_results_configuration
from expam.classify.taxonomy import TaxonomyNCBI
from expam.cli.main import CommandGroup, ExpamOptions
from expam.database import FileLocationConfig
from expam.database.config import JSONConfig, make_database_config, validate_database_file_configuration
from expam.sequences import format_name
from expam.tree.simulate import simulate_balanced_phylogeny
from expam.utils import die, ls


class UtilsCommand(CommandGroup):
    commands: Set[str] = {
        'download_taxonomy', 'cutoff', 'fake_phylogeny', 'plot_memory'
    }

    def __init__(
        self, config: FileLocationConfig, out_dir: str, cutoff: int, cpm: float
    ) -> None:
        super().__init__()

        self.config: FileLocationConfig = config
        self.out_dir: str = out_dir
        self.results_config: ResultsPathConfig = None if out_dir is None else make_results_config(out_dir)

        self.json_conf = None

        self.count_cutoff = cutoff
        self.cpm = cpm

    @classmethod
    def take_args(cls, args: ExpamOptions) -> dict:
        cutoff = cls.parse_ints(args.cutoff)
        cpm = cls.parse_floats(args.cpm)

        return {
            'config': make_database_config(args.db_name),
            'out_dir': args.out_url,
            'cutoff': cutoff,
            'cpm': cpm
        }

    def get_conf(self):
        if self.json_conf is None:
            self.check_database_exists()
            self.json_conf = JSONConfig(self.config.conf)
        
        return self.json_conf

    def check_database_exists(self):
        validate_database_file_configuration(self.config)

    """
    Download taxonomy command
    =========================
    
    """
    def download_taxonomy(self):
        self.check_database_exists()

        tax_obj: TaxonomyNCBI = TaxonomyNCBI(self.config)
        tax_obj.accession_to_taxonomy()

    """
    Employ cutoff on taxonomic classification output
    ================================================
    
    """
    def cutoff(self):
        if self.out_dir is None:
            die("Must supply -o/--out!")

        validate_results_configuration(self.results_config, check_taxonomy=True)

        if os.path.exists(self.out_dir):
            validate_classification_results(self.out_dir)
        else:
            raise ValueError("Results path %s doesn't exist!")

        file_names = [
            file_dir
            for file_dir in ls(self.results_config.tax)
            if file_dir != self.results_config.tax_classified and file_dir != self.results_config.tax_split
        ]

        for summary_file in file_names:
            print("Employing cutoff on sample %s..." % summary_file)

            with open(summary_file, "r") as f:
                data = [line.strip().split('\t') for line in f]

            total_classifications = sum(int(line[3]) for line in data[1:])
            total_splits = sum(int(line[6]) for line in data[1:])
            total_reads = total_classifications + total_splits

            if self.cpm is not None:
                cutoff = self.cpm * (total_reads / 1e6)
            elif cutoff is None:
                raise Exception("Require a cutoff value!")

            def valid_line(line):
                if int(line[2]) >= cutoff or int(line[5]) >= cutoff:
                    return True

            valid_taxids = [line[0] for line in data[1:] if valid_line(line)]
            valid_data = [line for i, line in enumerate(data) if i == 0 or valid_line(line)]

            # Overwrite summary.
            with open(summary_file, "w") as f:
                f.write("\n".join("\t".join(line) for line in valid_data))

            # Overwrite raw output.
            raw_sample_output: str = os.path.join(self.results_config.tax_raw, os.path.basename(summary_file))
            with open(raw_sample_output, "r") as f:
                valid_data = []

                for line in f:
                    current_data = line.strip().split('\t')
                    if current_data[2] in valid_taxids:
                        valid_data.append('\t'.join(current_data))

            with open(raw_sample_output, "w") as f:
                f.write('\n'.join(valid_data))

    """
    Generate a fake phylogeny
    =========================
    
    """
    def fake_phylogeny(self):
        conf: JSONConfig = self.get_conf()

        # Get list of sequences.
        seq_names = [
            format_name(os.path.basename(url))
            for url in conf.get_paths()
        ]

        if not seq_names:
            die("No sequences added to the database!")

        # Compute balanced phylogeny.
        fake_phy_path = os.path.join(self.config.phylogeny, 'expam_outtree.nwk')
        newick_str = simulate_balanced_phylogeny(seq_names)

        with open(fake_phy_path, "w") as f:
            f.write(newick_str)

        print("Newick file generated at %s." % fake_phy_path)

        # Add phylogeny to expam configuration file.
        conf.set(phylogeny_path=fake_phy_path)
        conf.save()

        print("Database phylogeny set to %s." % fake_phy_path)

    """
    Plot expam database build memory usage
    ======================================
    
    """
    def plot_memory(self):
        conf: JSONConfig = self.get_conf()

        def _get_mem(data: str):
            start, end = data.find("(") + 1, data.find(")")

            if start == -1 or end == -1:
                raise ValueError("Invalid data line %s!" % data)

            rss_string, vsz_string = data[start:end].split(", ")

            for point in (rss_string, vsz_string):
                _, mem_used = point.split(" ")

                yield float(mem_used.replace("Gb", ""))

        def _r_yield_dict(obj, current_key=""):
            if isinstance(obj, dict):
                for key, value in obj.items():
                    yield from _r_yield_dict(value, "%s_%s" % (current_key, str(key)))
            else:
                yield current_key, obj

        job_types = {
            "_extract.log": "extract",
            "_union.log": "union",
        }
        plot_colours = {
            "extract": {
                "rss": "blue",
                "vsz": "black",
            },
            "union": {
                "rss": "green",
                "vsz": "purple"
            }
        }
        job_labels = {
            "extract": "extract",
            "union": "union",
        }

        log_urls = [
            os.path.join(self.config.logs, log_name)
            for log_name in ls(self.config.logs, ext=".log")
            if "_main" not in log_name
        ]

        _, t0 = self.get_t0(self.config.logs)

        # Collect total memory usage as well.
        cumulative = {
            "union": {
                "rss": 0,
                "vsz": 0,
            },
            "extract": {
                "rss": 0,
                "vsz": 0,
            }
        }

        fig, (memory_time_ax, cumulative_memory_ax) = plt.subplots(1, 2, figsize=(20, 10))

        for log_url in log_urls:

            with open(log_url, "r") as f:
                log_data = f.readlines()

            t, rss, vsz = [], [], []

            for line in log_data:
                if "of RAM" in line:
                    rss_mem, vsz_mem = _get_mem(line)
                    time_point = (self.get_time(line) - t0).total_seconds()

                    rss.append(rss_mem)
                    vsz.append(vsz_mem)
                    t.append(time_point / 3600)  # Time in hours.

            for mem_type, mem_data in (("rss", rss), ("vsz", vsz)):
                for k, v in job_types.items():
                    if k in log_url:
                        job_type = v

                        break
                else:
                    raise TypeError("Unknown process log type %s!" % os.path.basename(log_url))

                plot_label = job_labels[job_type]
                # Don't allow duplicate labels.
                job_labels[job_type] = ""

                memory_time_ax.plot(
                    t,
                    mem_data,
                    label=plot_label,
                    marker=".",
                    color=plot_colours[job_type][mem_type]
                )

                if mem_data:
                    # Add to cumulative memory usage.
                    cumulative[job_type][mem_type] += mem_data[-1]

        # Memory-over-time axis labels.
        memory_time_ax.set(
            xlabel="Time since start [hrs]",
            ylabel="Memory used [Gb]",
            title="Memory used by child workers"
        )
        memory_time_ax.legend(loc="best")

        # Bar chart of cumulative memory usage.
        labels = ["union", "extract"]
        rss = [cumulative["union"]["rss"], cumulative["extract"]["rss"]]
        vsz = [cumulative["union"]["vsz"], cumulative["extract"]["vsz"]]

        # Set axis positions.
        x = list(range(len(labels)))
        width = 0.35

        rss_x = list(v - width / 2 for v in x)
        vsz_x = list(v + width / 2 for v in x)

        # Plot bar charts.
        rss_bars = cumulative_memory_ax.bar(rss_x, rss, width, label="RSS")
        vsz_bars = cumulative_memory_ax.bar(vsz_x, vsz, width, label="VSZ")

        # Cumulative memory usage axis labels.
        cumulative_memory_ax.set(
            ylabel="Total memory used [Gb]",
            xticks=x,
            xticklabels=labels,
            title="Total memory used by workers",
        )
        cumulative_memory_ax.legend(loc="best")

        # Save figure to disk.
        fig.savefig(os.path.join(self.config.logs, "memory_usage.png"))

    


from multiprocessing.shared_memory import SharedMemory
from os import ftruncate


class ResizableSharedMemory(SharedMemory):
    def __init__(self, name=None, create=False, size=0):
        super(ResizableSharedMemory, self).__init__(name, create, size)

    @property
    def size(self):
        """Size in bytes."""
        return self._size

    @size.setter
    def size(self, val):
        self._size = val

    def resize(self, newsize):
        fd = self._fd

        try:
            ftruncate(fd, newsize)
            return 1, newsize
        except OSError as e:
            return 0, e

    def remap(self, newsize):
        if newsize != self.size:
            del self._buf  # Clear buffer exposure.
            self._mmap.resize(newsize)  # Remap memory.
            self._buf = memoryview(self._mmap)  # Re-expose buffer.
            # Update size.
            self.size = newsize




import os

from expam.classify import PHY_CLASSIFIED_FILE, PHY_RAW, PHY_RESULTS, PHY_SPLIT_FILE, TAX_CLASSIFIED_FILE, TAX_RAW, TAX_RESULTS, TAX_SPLIT_FILE, TEMP_RESULTS, ResultsPathConfig
from expam.utils import die


def make_results_config(out_path: str) -> ResultsPathConfig:
    output_file_locations = {
        'base': out_path,
        'phy': os.path.join(out_path, PHY_RESULTS),
        'tax': os.path.join(out_path, TAX_RESULTS),
        'temp': os.path.join(out_path, TEMP_RESULTS),
        'phy_raw': os.path.join(out_path, PHY_RAW),
        'tax_raw': os.path.join(out_path, TAX_RAW),
        'phy_classified': os.path.join(out_path, PHY_CLASSIFIED_FILE),
        'phy_split': os.path.join(out_path, PHY_SPLIT_FILE),
        'tax_classified': os.path.join(out_path, TAX_CLASSIFIED_FILE),
        'tax_split': os.path.join(out_path, TAX_SPLIT_FILE)
    }
    return ResultsPathConfig(**output_file_locations)

def load_results_config(out_path: str, create: bool = False) -> ResultsPathConfig:
    proposed_config: ResultsPathConfig = make_results_config(out_path)

    # Make base results path.
    if create:
        if not os.path.exists(out_path):
            try:
                os.mkdir(out_path)
            except OSError:
                print("Failed to make results path %s." % out_path)

        create_results(proposed_config)
    elif not validate_results_configuration(proposed_config, check_taxonomy=False):
        die("Results path does not exist!")
        
    return proposed_config

def create_results(config: ResultsPathConfig):
    for path_field in ('phy', 'phy_raw', 'temp'):
        path = getattr(config, path_field)

        if not os.path.exists(path):
            os.mkdir(path)

def create_tax_results(config: ResultsPathConfig):
    for path_field in ('tax', 'tax_raw'):
        path = getattr(config, path_field)

        if not os.path.exists(path):
            os.mkdir(path)

def validate_results_configuration(config: ResultsPathConfig, check_taxonomy: bool = True):
    phy_files = (config.phy, config.phy_classified, config.phy_split)

    for phy_file in phy_files:
        if not os.path.exists(phy_file):
            return False
    
    return True

def validate_classification_results(results_dir: str):
    if not os.path.exists(results_dir):
        die("Could not find results %s!" % results_dir)

    results_config: ResultsPathConfig = make_results_config(results_dir)

    if not (os.path.exists(results_config.phy_classified) or os.path.exists(results_config.phy_split)):
        raise Exception("Path does not look like expam results folder!")

